#!/usr/bin/env python
# Time-stamp: <2003-03-31 11:16:43 crabbkw>
# Code and design by Casey Crabb (crabbkw@nafai.dyndns.org)
# This code is licensed under the BSD license.
# See the LICENSE file for details
#
# Copyright Casey Crabb (crabbkw@nafai.dyndns.org) July 2001
#

# from __future__ import nested_scopes
from threading import *
from IMCom import IMCom, errorCodes
from Preferences import Preferences
from LogHandler import LogHandler
from AutoStatus import AutoStatus
from CICommands import *
from Colors import *
import string
import socket
import os
import time
import sys
import getpass
import operator
import signal
import traceback
import SocketWrapper
import pprint

try:
    import codecs
    CODECS = 1
except:
    CODECS = 0

try:
    import locale
    LOCALE = 1
except:
    LOCALE = 0

try:
    import readline
    READLINE = 1
except:
    READLINE = 0


VERSION = "1.30beta8"

terminalWidth = 80
pp = pprint.PrettyPrinter(indent=4)

def mySort(x, y):
    a,b = x
    c,d = y
    return cmp(b,d)

def getTerminalWidth():
    terminalWidth = 80 #linelength

class CLI(Thread):
    "The Main thread and heart of the IMCom CLI interface."
    def __init__(self):
        Thread.__init__(self)
        self.VERSION = VERSION
        self.threadHash = {}
        self.prompt = "IMCom> "
        self.lastMessaged = None
        self.lastReceived = None
        self.lastToDict = {}
        self.lastFromDict = {}
        self.showPresenceUpdates = 1
        self.ringBell = 1
        self.prefs = Preferences()
        self.profile = self.prefs.getDefaultProfile()
        self.imcom = IMCom(self)
        self.registerCallbacks()
        self.loghandler = LogHandler( self, self.profile )
        self.currentStatus = "online"
        self.currentStatusType = "auto"
        self.currentStatusReason = ""
        self.debug = 0
        self.debugdetails = {
            "autostatus" : 0,
            }
        self.mode = 0
        self.MULTILINE = 1
        self.outputQueue = []
        self.gettingCommand = 0
        self.pendingCommand = 0
        self.userQueue = []
        self.commandQueue = []
        self.socketlisteners = []
        self.socketserver = ''
        self.autostatus = None

        if(READLINE):
            readline.parse_and_bind("tab: complete")
            readline.parse_and_bind("set bell-style none")
            readline.set_completer_delims(" ")
            readline.set_completer(tabCompleter)


    def registerCallbacks(self):
        self.imcom.cbHandleConferenceMessage  = self.handleConferenceMessage
        self.imcom.cbHandleDisconnected       = self.handleDisconnected
        self.imcom.cbHandleAdminWho           = self.handleAdminWho
        self.imcom.cbHandlePresenceUpdate     = self.handlePresenceUpdate
        self.imcom.cbHandlePresenceError      = self.handlePresenceError
        self.imcom.cbHandleMessageReceive     = self.handleMessageReceive
        self.imcom.cbHandleMessageError       = self.handleMessageError
        self.imcom.cbHandleIQError            = self.handleIQError
        self.imcom.cbHandleInfoError          = self.handleInfoError
        self.imcom.cbHandleFileReceive        = self.handleFileReceive
        self.imcom.cbHandleFileReceived       = self.handleFileReceived
        self.imcom.cbHandleFileErrorReceived  = self.handleFileErrorReceived
        self.imcom.cbHandleAgentList          = self.handleAgentList
        self.imcom.cbHandleAgentRegister      = self.handleAgentRegister
        self.imcom.cbHandleAgentRegistered    = self.handleAgentRegistered
        self.imcom.cbHandleSubscribe          = self.handleSubscribe
        self.imcom.cbHandleSubscribed         = self.handleSubscribed
        self.imcom.cbHandleUnsubscribed       = self.handleUnsubscribed
        self.imcom.cbHandleUnsubscribe        = self.handleUnsubscribe
        self.imcom.cbHandleRosterUpdateCheck  = self.handleRosterUpdateCheck
        self.imcom.cbHandleRosterUpdate       = self.handleRosterUpdate
        self.imcom.cbHandleVCardSubmit        = self.handleVCardSubmit
        self.imcom.cbHandleVCard              = self.handleVCard
        self.imcom.cbHandleNoVCard            = self.handleNoVCard
        self.imcom.cbHandleLogin              = self.login
        self.imcom.cbHandleGrabRoster         = self.grabRoster
        self.imcom.cbHandleAgentUnRegistered  = self.handleAgentUnRegistered
        self.imcom.cbHandleNegotiationRequest = self.handleNegotiationRequest
        self.imcom.cbHandleNegotiationResult  = self.handleNegotiationResult
        self.imcom.cbHandleVersionResponse    = self.handleVersionResponse
        self.imcom.cbHandleChangePassword     = self.handleChangePassword
        self.imcom.cbHandleConferenceMessage  = self.handleConferenceMessage
        self.imcom.cbHandleConferencePresence = self.handleConferencePresence
        self.imcom.cbHandleConferenceCreated  = self.handleConferenceCreated
        self.imcom.cbHandleConferenceNicknameChange = self.handleConferenceNicknameChange
        self.imcom.cbHandleConferenceNicknameChangeSuccess = self.handleConferenceNicknameChangeSuccess
        self.imcom.cbHandleConferenceInvite   = self.handleConferenceInvite
        self.imcom.cbHandleConferenceSubject = self.handleConferenceSubject
        self.imcom.cbHandleConferenceKicked = self.handleConferenceKicked
        self.imcom.cbHandleConferenceUserKicked = self.handleConferenceUserKicked
        self.imcom.cbHandleConferenceBanned = self.handleConferenceBanned
        self.imcom.cbHandleConferenceUserBanned = self.handleConferenceUserBanned
        self.imcom.cbHandleConferenceDestroyed = self.handleConferenceDestroyed
        self.imcom.cbHandleConferenceConfig = self.handleConferenceConfig
        self.imcom.cbHandleConferenceConfigSet = self.handleConferenceConfigSet
        self.imcom.cbHandleNickCollision = self.handleNickCollision

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # -------------------------------------------------------------------------
    #  Functions for the login procedure
    # -------------------------------------------------------------------------
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def grabRoster(self):
        """This function is called by the IMCom library when the roster
        has been downloaded. Because the presence events for people
        can come before the roster is downloaded this function needs
        to print out people for whom we've already gotten presence
        updates from.  This is because the presence code doesn't show
        us presence information for anyone who is not on our roster."""
        self.output("Roster has been downloaded")
        keys = self.imcom.jidhash.keys()
        for item in keys:
            if(self.imcom.preshash.has_key(item)):
                available, status, show = self.imcom.preshash[item]
                if(not available):
                    continue
                if(self.showPresenceUpdates):
                    if( self.imcom.gjidhash.has_key(item) and
                        (operator.contains(self.imcom.gjidhash[item],
                                           "lurker") or
                         operator.contains(self.imcom.gjidhash[item],
                                           "ignore"))):
                        return
                    a,show,status = self.imcom.preshash[item]
                    nick = self.getNick(item)
                    ttime = self.getTime()
                    ddate = self.getDate()
                    self.output(usercolor + nick + defaultcolor +
                                " changed status to " +
                                statuscolor + show + defaultcolor + " (" +
                                desccolor + status + defaultcolor + ") at " +
                                timecolor + ttime + defaultcolor)
            else:
                self.imcom.preshash[item] = (0, "unavailable", "offline")

    def login(self, successful):
        """This function is called by the IMCom library. It lets us
        know whether the login was successful or not."""
        if(successful):
            self.output(usercolor + "Logged on" + defaultcolor)
        else:
            self.output(errorcolor + "Login FAILED!" + defaultcolor)

    def applyProfileWrapper(self):
      result = (-1, "No Error")
      iter = 0
      for iter in range(5):
        result = cli.applyProfile()
        if result[0] == 0:
          return result

        print result[1]
        if result[0] == 1:
          sys.exit(-1)
        print "Sleeping for 5 seconds."
        time.sleep(5)

      print "Giving up."
      return result

    def applyProfile(self):
        """This function actually begins the login process by
        selecting a profile and calling the IMCom library's
        changeProfile command.  If no profile is available we allow
        the user to create one. This does not let the user create an
        account, only lets them to create a profile for an account
        pre-existing."""
        if(self.profile!=None):
            p = self.profile
            useSSL = 1
            if(p.switches['ssl'][0] == "false" or
               p.switches['ssl'][0] == "no" or
               p.switches['ssl'][0] == "off" or
               p.switches['ssl'][0] == 0):
              useSSL = 0

            self.imcom.DTCPPORT = p.DTCPPort
            result = self.imcom.changeProfile(p.server, p.port, p.user, p.password, p.resource, useSSL, p.priority, p.encoding)
            if result[0] != 0:
              return result

            self.loghandler.initLogHandler(p)
            self.autostatus = AutoStatus( self, self.profile )
            self.autostatus.start()
            self.loghandler.setProfile( self.profile )
            self.autostatus.setProfile( self.profile )
            return result
        else:
            # I should create a new profile here.
            print("No profiles available or default profile not set.")
            self.createNewProfile()
            self.profile = self.prefs.getDefaultProfile()
            return self.applyProfile()

    def applyPrefs(self):
        """This function sets all the preference switches."""
        if(self.profile == None):
            return
        p = self.profile
        if(p.switches['ringbell'][0] == "true"):
            self.ringBell = 1
        else:
            self.ringBell = 0

        if( p.switches['colors'][0] == "true"):
            self.setColors()
        else:
            self.setNoColors()

        if( p.switches['debug'][0] == "false" ):
            self.debug = 0
            self.imcom.setDebug(0)
        else:
            self.debug = 1
            self.imcom.setDebug(1)

        if( p.switches['nickprompt'][0] == "false"):
            self.useNickAsPrompt = 0
        else:
            self.useNickAsPrompt = 1

        if( p.switches['allowinterrupt'][0] == "false" ):
            self.allowInterrupt = 0
        else:
            self.allowInterrupt = 1

        if( p.switches['statusshow'][0] == "false" or
        p.switches['statusshow'][0] == "no"):
            self.showPresenceUpdates = 0
        else:
            self.showPresenceUpdates = 1

    def createNewProfile(self):
        """This function creates a new profile and overwrites the
        preferences file ~/.imcom/imcomrc"""
        print("You don't seem to have a profile, lets make a new one")
        name = raw_input("Enter a profile name > ")
        server = raw_input("Enter the server name > ")
        port = raw_input("Enter the port (5222) > ")
        if(port == None or len(port) == 0):
            port = "5222"
        user = raw_input("Enter your login > " )
        password = getpass.getpass("Enter your password > ")
        resource = raw_input("Enter the resource (imcom) > ")
        if(resource == None or len(resource) == 0):
            resource = "imcom"
        encoding = raw_input("Enter the encoding you want to use (iso-8859-1) > ")
        if(encoding == None or len(encoding) == 0):
            encoding = "iso-8859-1"
        self.prefs.createInitialProfile(name,server,port,
                                        user,password,resource, encoding)


    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # -------------------------------------------------------------------------
    #  The functions which begin with handle are callbacks from the
    #  IMCom library
    # -------------------------------------------------------------------------
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

    def handleDisconnected(self):
        """This is called by the IMCom library when the server has
        disconnected us unexpectantly. In theory it should go into
        EVAL mode essentially dumped a developer onto a Python console
        with the state of IMCom preserved. This however, does not work
        well yet."""
        self.output(errorcolor + "Server Disconnected us at " +
                    timecolor + self.getTime() + defaultcolor)
        if(self.debug):
            self.output(errorcolor + "Going into EVAL mode, EOF (Control-d) to quit." + defaultcolor)
            line = "blah"
            while line:
                try:
                    line = raw_input("> ")
                    try:
                        pp.pprint(eval(line))
                    except:
                        print errorcolor + "That didn't eval well" + defaultcolor
                except EOFError:
                    sys.exit(-1)
                except KeyboardInterrupt:
                    sys.exit(-1)
            sys.exit(-1)
        self.imcom = None
        self.imcom = IMCom(self)
        self.registerCallbacks()

        result = self.applyProfileWrapper()
        if result[0] != 0:
          sys.exit(-1)

        if(self.currentStatus == "chat"):
            self.imcom.sendChat()
        elif(self.currentStatus == "away"):
            self.imcom.sendAway(self.currentStatusReason)
        elif(self.currentStatus == "xa"):
            self.imcom.sendXA(self.currentStatusReason)
        elif(self.currentStatus == "dnd"):
            self.imcom.sendDND(self.currentStatusReason)



    def handleAdminWho(self, who, show, status, available, resource):
        """This function is called by the IMCom library when there has been a request for the list of users on the server. You
        must be admin for this function to work."""
        s = usercolor + who + "/" + resource + defaultcolor + " is " + statuscolor + show + defaultcolor + " as " + desccolor + \
            status + defaultcolor
        self.output(s)


    def handlePresenceUpdate(self, who, show, status, available, resource,
        duplicate):
        """This function is called by the IMCom library when there has
        been a change in someone's precense. The user must be
        subscribed to the presence."""
        #if(self.debug):
        #    logDebug("Updating presence of " + who + " to " + show +
        #             " as " + status)

        # don't print the status of anyone not in our jidlist and not
        # a transport
        #if( (not self.imcom.jidhash.has_key(who) or
        #     not self.imcom.gjidhash.has_key(who) ) and
        #    ( string.find( who, "@" ) != -1 ) ):
        #    if(self.debug):
        #        logDebug("jidhash doesn't have a " + who)
        #    return

        nick = self.getNick(who)

        # these handle error conditions in transports.
        if( ( string.find( who, "@" ) == -1 )
            and show and show == "offline" ):
            s = errorcolor + "ERROR: " + defaultcolor + \
                nick + " is offline"
            if(status):
                s = s + " : " + status
            self.output(s)
            return
        if( ( string.find( who, "@" ) == -1 )
            and show and ( string.find( show, " " ) != -1 ) ):
            s = errorcolor + "Notice: " + defaultcolor + \
                nick + ": " + show
            self.output(s)
            return

        # this is in response to a bug in trillian (icq program)
        # yes, the fix should be server-side but you know....
        if( ( self.profile.switches['igndupstatus'][0] == "true" ) 
            and duplicate ):
            return

        # commented out for testing/reimplementation of how IMCom
        # deals with transports.
        #if( string.find( who, "@" ) == -1 and not self.imcom.agenthash.has_key(who)):
        #    return

        # don't display the presence of people in the lurker group.
        if( self.imcom.gjidhash.has_key(who) and
            (operator.contains(self.imcom.gjidhash[who],"lurker") or
             operator.contains(self.imcom.gjidhash[who],"ignore"))):
            if(self.debug):
                logDebug("Skipping lurker: " + who)
            return

        # only display status updates if we are set to display status updates.
        if(self.showPresenceUpdates):
            ttime = self.getTime()
            ddate = self.getDate()
            if(available):
                s = usercolor + nick + defaultcolor
                # if they have a resource, and they're not a transport
                # then add the resource string.
                if(resource != None and resource != "" and not
                   self.imcom.agenthash.has_key(who)):
                    s = s + "/" + keycolor + resource + defaultcolor

                # append the rest of the information
                s = s + " changed status to " + \
                    statuscolor + show + defaultcolor + " (" + \
                    desccolor + status + defaultcolor + ") at " + \
                    timecolor + ttime + defaultcolor

                self.output(s)
            else:
                s= usercolor + nick + defaultcolor
                # if they have a resource, and they're not a transport
                # then add the resource string.
                if(resource != None and resource != "" and not
                   self.imcom.agenthash.has_key(who)):
                    s = s + "/" + keycolor + resource + defaultcolor

                # append the rest of the information
                s = s + " is unavailable : " + \
                    desccolor + status + defaultcolor + \
                    " at " + timecolor + ttime + defaultcolor
                self.output(s)

    def handlePresenceError(self, ffrom, code, text):
      towrite = errorcolor + "ERROR: " + usercolor + ffrom + defaultcolor + " says " + text + "(" + code + ")."
      if self.imcom.conferences.has_key(ffrom):
        del self.imcom.confnick[self.imcom.conferences[ffrom][0]]
        del self.imcom.conferences[ffrom]
      self.output(towrite)
      return


    def handleConferenceMessage(self, conf, nick, body, delay):
        "This function is called from the IMCom library when a message has been received from a conference"
        if(self.profile.switches['confsuppress'][0] == "false"):
            towrite = keycolor + conf + defaultcolor  + "/" + usercolor + nick + defaultcolor
            ttime = self.getTime()
            ddate = self.getDate()
            towrite = towrite + " - Conference Message - "

            # append the timestamp and the rest of the information
            towrite = towrite + timecolor + ddate + " |  " + ttime + defaultcolor
            towrite = towrite + "\n" + messagebodycolor + body + defaultcolor
        else:
            towrite = keycolor + conf + defaultcolor + "/" + usercolor + nick + defaultcolor  + " : "
            towrite = towrite + messagebodycolor + body + defaultcolor

        # ring the terminal bell if configured to
        if(self.ringBell):
            self.printText(chr(7))

        self.output(towrite)


    def handleConferenceCreated(self, conf):
      towrite = usercolor + conf + defaultcolor + " was just created by you. You need to set it up. I am requesting the form."
      self.output(towrite)
      self.imcom.sendConferenceConfigRequest(conf)
      return

    def handleConferenceConfigSet(self, conf):
      towrite = usercolor + conf + defaultcolor + " configuration was set successfully."
      self.output(towrite)
      return

    def handleConferenceConfig(self, conf, xmlform):
      towrite = messagebodycolor + xmlform.instructions + defaultcolor
      self.output(towrite)
      answers = self.getXDataAnswers(xmlform.fields)
      self.imcom.sendConferenceConfigResults(conf, answers)

    def handleConferenceNicknameChange(self, conf, oldnick, newnick):
      towrite = keycolor + self.imcom.conferences[conf][0] + defaultcolor + "/" + usercolor + oldnick + defaultcolor
      towrite = towrite + " has changed their nickname to " + usercolor + newnick + defaultcolor + "."
      self.output(towrite)

    def handleConferenceNicknameChangeSuccess(self, conf, oldnick, newnick):
      towrite = "You successfully changed your nick name from " + usercolor + oldnick + defaultcolor
      towrite = towrite + " to " + usercolor + newnick + defaultcolor + " in conference "
      towrite = towrite + keycolor + self.imcom.conferences[conf][0] + defaultcolor + "."
      self.output(towrite)

    def handleConferencePresence(self, conf, nick, show, reason, affiliation, role, jid):
      "This function is called from the IMCom library when someone in a conference has changed status"
      if(self.showPresenceUpdates):# and 0):
        if reason == None:
          reason = show
        towrite = keycolor + self.imcom.conferences[conf][0] + defaultcolor + "/" + usercolor + nick + defaultcolor
        towrite = towrite + " changed status to " + statuscolor + show + defaultcolor
        towrite = towrite + " as " + desccolor + reason + defaultcolor
        if affiliation != None:
          towrite = towrite + " (" + keycolor + affiliation + defaultcolor + ") "
        if role != None:
          towrite = towrite + " (" + keycolor + role + defaultcolor + ") "
        ttime = self.getTime()
        towrite = towrite + " at " + timecolor + ttime + defaultcolor
        self.output(towrite)

    def handleConferenceInvite(self, conf, ffrom, reason, password):
      towrite = usercolor + ffrom + defaultcolor + " invites you to join " + keycolor + conf + defaultcolor + "\n"
      towrite = towrite + " -- " + reason
      if password != None:
        towrite = towrite + "\nThe password is: " + keycolor + password + defaultcolor
      self.output(towrite)

    def handleConferenceSubject(self, conf, ffrom, subject):
      towrite = keycolor + conf + defaultcolor + "/" + usercolor + ffrom + defaultcolor + " set the subject to "
      towrite = towrite + desccolor + subject + defaultcolor
      self.output(towrite)

    def handleConferenceKicked(self, conf, reason, jidOfKicker):
      towrite = errorcolor + "KICKED: " + defaultcolor + " from " + keycolor + conf + defaultcolor
      if jidOfKicker != None:
        towrite = towrite + " by " + usercolor + jidOfKicker + defaultcolor
      if reason != None:
        towrite = towrite + " : " + reason
      self.output(towrite)

    def handleConferenceUserKicked(self, conf, person, reason, jidOfKicker):
      towrite = keycolor + self.imcom.conferences[conf][0] + defaultcolor + "/" + usercolor + person + defaultcolor
      towrite = towrite + " was kicked"
      if jidOfKicker != None:
        towrite = towrite + " by " + usercolor + jidOfKicker + defaultcolor
      if reason != None:
        towrite = towrite + " for : " + desccolor + reason + defaultcolor
      towrite = towrite + "."
      self.output(towrite)

    def handleConferenceBanned(self, conf, reason, jidOfKicker):
      towrite = errorcolor + "BANNED: " + defaultcolor + " from " + keycolor + conf + defaultcolor
      if jidOfKicker != None:
        towrite = towrite + " by " + usercolor + jidOfKicker + defaultcolor
      if reason != None:
        towrite = towrite + " : " + reason
      self.output(towrite)

    def handleConferenceUserBanned(self, conf, person, reason, jidOfKicker):
      towrite = keycolor + self.imcom.conferences[conf][0] + defaultcolor + "/" + usercolor + person + defaultcolor
      towrite = towrite + " was banned"
      if jidOfKicker != None:
        towrite = towrite + " by " + usercolor + jidOfKicker + defaultcolor
      if reason != None:
        towrite = towrite + " for : " + desccolor + reason + defaultcolor
      towrite = towrite + "."
      self.output(towrite)

    def handleConferenceDestroyed(self, conf):
      towrite = keycolor + conf + defaultcolor + " has been " + errorcolor + "destroyed" + defaultcolor + "."
      self.output(towrite)

    def handleMessageReceive(self, ffrom, body, thread, delay,
                             resource):
        """This function is called from the IMCom library when a
        message has been received from a user or transport/service."""

        # don't display messages from people who belong to the ignore
        # group.
        if( self.imcom.gjidhash.has_key(ffrom) and
            operator.contains(self.imcom.gjidhash[ffrom],"ignore")):
            if(self.debug):
                logDebug("Skipping message from ignored user: " + ffrom)
            return

        # append the jid to the top of the stack of last users to
        # event us.
        self.appendUserQueue(ffrom)

        # try and get their nickname. otherwise their nickname is
        # their jid.
        try:
            nick = self.imcom.jidhash[ffrom]
        except:
            nick = ffrom

        ttime = None
        ddate = None
        towrite = None
        towrite = usercolor + nick + defaultcolor

        if(resource != ""):
            towrite = towrite + "/" + keycolor + resource + defaultcolor

        # if the delay timestamp exists, format it properly, otherwise
        # the current time.
        if(delay != None):
            # parse the delay, attach the timestamp in good format.
            # tuple is (year, month, day, hour, minutes, seconds,
            #           daysintotheyear, dst)
            mytup = (int(delay[:4]),int(delay[4:6]),int(delay[6:8]),
                     int(delay[9:11]),int(delay[12:14]),int(delay[15:17]),0,0,0)
            thet = time.localtime( time.mktime(mytup) - time.timezone)
            ttime = time.strftime('%H:%M:%S',thet)
            ddate = time.strftime('%m/%d/%Y',thet)
            towrite = towrite + " - " + errorcolor + \
                      "DELAYED " + defaultcolor + "Message - "
        else:
            ttime = self.getTime()
            ddate = self.getDate()
            towrite = towrite + " - Message - "

        # append the timestamp and the rest of the information
        towrite = towrite + timecolor + ttime + defaultcolor
        if(delay != None):
            towrite = towrite + " on " + timecolor + ddate
        towrite = towrite + "\n" + messagebodycolor + body + defaultcolor

        # ring the terminal bell if configured to
        if(self.ringBell):
            if(self.profile.audiocmd != None):
                os.spawnv(os.P_NOWAIT,self.profile.audiocmd,(self.profile.audiocmd,self.profile.audioargs,))
            self.printText(chr(7))

        self.lastReceived = ffrom
        self.output(towrite)
        # threadhash is used so that if someone is seeing this as a
        # particular conversation we continue the conversation. (Silly
        # GUI apps)
        self.threadHash[ffrom] = thread
        self.logMessage(ffrom,ffrom,ddate,ttime,body)



    def handleMessageError(self, ffrom, body, code):
        """This function is called by the IMCom library when there has
        been some sort of error related to sending or receiving a
        message. CLI just prints out the information for the user."""
        if( ( code == "0" or code == 0 ) and body != "" ):
            towrite = errorcolor + "ERROR: " + defaultcolor + body
        else:
            towrite = errorcolor + "ERROR: " + defaultcolor + \
                      "sending message to"\
                      " " + usercolor + ffrom + defaultcolor + " :"\
                      " " + errorcolor + errorCodes[code] + defaultcolor
        self.output(towrite)



    def handleIQError(self, ffrom, code, body):
        """This function is called by the IMCom library when there has
        been some sort of error related to information request or
        submission. CLI just prints out the information for the user."""
        towrite = errorcolor + "ERROR: " + defaultcolor + "InfoQuery event"\
                  " " + usercolor + ffrom + defaultcolor + " :"\
                  " " + errorcolor + errorCodes[code] + defaultcolor
        if body != None:
            towrite = towrite + " : " + errorcolor + body + defaultcolor
        self.output(towrite)


    def handleChangePassword(self, success, newPassword, errorCode, errorText):
      towrite = None
      if success == 1:
        towrite = "Your password has been successfully change."
        self.profile.password = newPassword
        self.prefs.writePreferences()
      else:
        towrite = errorcolor + "ERROR: " + defaultcolor + " Your password could not be changed."
        towrite = towrite + errorcolor + errorCodes[code] + defaultcolor + errorText
      self.output(towrite)
      return

    def handleInfoError(self, code, text):
        """This function is called by the IMCom library when there has
        been some sort of error related to information request.
        CLI just prints out the information for the user."""
        towrite = errorcolor + "ERROR: " + defaultcolor + "getting info"\
                  " Code: " + errorcolor + code + defaultcolor + " "\
                  " Text: " + errorcolor + text + defaultcolor
        self.output(towrite)



    def handleFileReceive(self, jid, url):
        """This function is called by the IMCom library when someone
        has requested to send us a file."""
        self.appendUserQueue(jid)
        nick = self.getNick(jid)
        towrite = usercolor + nick + defaultcolor + " wants to send you"\
                  " a file at URL:\n" + keycolor + url + defaultcolor +\
                  "\nTo receive the file type: \n" + \
                  self.prefs.getCommand(self.profile, "getfile") + \
                  " " + nick
        self.output(towrite)



    def handleFileReceived(self, jid, url, id):
        """This function is called by the IMCom library when we have
        successfully received the file."""
        self.appendUserQueue(jid)
        nick = self.getNick(jid)
        file = url[url.rfind('/')+1:]
        towrite = "Successfully receieved a file from " + usercolor +\
                  nick + defaultcolor + \
                  "\nThe file is saved in " + keycolor + "~/.imcom/files/" + \
                  file + defaultcolor
        self.output(towrite)



    def handleFileErrorReceived(self, jid, url, id, type, text):
        """This function is called by the IMCom library when there has
        been some sort of error in file transmission."""
        self.appendUserQueue(jid)
        towrite = errorcolor + "Error: " + defaultcolor + \
                  "An error occured trying to get a file from " + \
                  usercolor + jid + defaultcolor + ".\nThe URL: " + \
                  keycolor + url + defaultcolor + "\n" + \
                  "Errortype: " + str(type) + "\n" + \
                  "Errorvalue: " + str(text)
        self.output(towrite)



    def handleAgentList(self, theList):
        """This function is called by the IMCom library when we have
        received a list of agents from the server. Agents is another
        word for Transport or Server in Jabber terminology."""
        self.output("The following is a list of transports available on" +
                    " your Jabber server.")
        for i in theList:
            if(i.name):
                self.output(keycolor + i.jid + defaultcolor +
                            " : " + defaultcolor + i.name)


    def handleAgentRegister(self, id, ffrom, instructions, fields):
        #
        # this will cause issues if gettingCommand is already set to 1
        # however, in that case things get *really* complicated anyway...
        # and it's not likely to happen unless the user is playing with us
        #
        self.output("Please finish whatever command you're currently entering (or just hit enter) to begin registration process.")
        self.pendingCommand = 1
        while(self.gettingCommand == 1):
          time.sleep(0.01)

        self.output("\n\nThe following are instructions to complete the "\
                    "registration process, they may be of use, then "\
                    "again, don't count on it.\n\n")
        self.output(desccolor + instructions + defaultcolor + "\n\n")
        self.gettingCommand = 1

        params = ffrom
        for i in fields:
            params += "<" + i

        interactiveregCommand(self, params)
        self.dumpQueue()
        self.pendingCommand = 0
        self.gettingCommand = 0


    def handleAgentRegistered(self, ffrom):
        """This function is called by the IMCom library when an agent
        has been successfully registered."""
        self.output("Your registration request to " + ffrom +
                    " was successful. I will authorize the transport now "
                    "for you.")
        addcommand = self.prefs.getCommand( self.profile, "add" )
        self.output( "To add contacts who use the " + self.getNick( ffrom ) +
            " use the command:\n" + addcommand +
            " id@" + ffrom + " nick\n" +
            "for instance, " + addcommand + " airog@" + ffrom + " airog" )
        self.imcom.sendSubscribed(ffrom) # + "/registered") is it needed?
        return

    def handleAgentUnRegistered(self, ffrom):
      """This function is called by the IMCom library when an agent has been successfully removed."""
      self.output("Your registration to " + ffrom + " was cancelled successfully.")
      return

    def handleSubscribe(self, ffrom, ask):
        """This function is called by the IMCom library when someone
        is requesting to add us to their roster."""
        self.appendUserQueue(ffrom)
        if(self.debug):
            logDebug("Handling a subscribe request, from : " + ffrom)
        ffrom = self.getNick(ffrom)
        self.output(usercolor + ffrom + defaultcolor +
                    " requests permission to subscribe to your presence")

    def handleSubscribed(self, ffrom):
        """This function is called by the IMCom library when someone
        authorized you to add them to your roster."""
        self.appendUserQueue(ffrom)
        nick = self.getNick(ffrom)
        self.output(usercolor + nick + defaultcolor +
                    " authorized your subscription request")

    def handleUnsubscribed(self, ffrom, success):
        """This function is called by the IMCom library when the
        unsubscribe process is complete."""
        self.appendUserQueue(ffrom)
        if(success):
            nick = self.getNick(ffrom)
            if self.imcom.asks[ffrom] == 'subscribe':
              self.output(usercolor + nick + defaultcolor +
                          " refused to grant authorization for you to subscribe to their presence. ")
              return
            self.output("Unsubscribed from " + usercolor + nick + defaultcolor +
                        "'s presence")
        else:
            self.output("Yowza! Unsubscription process failed!")


    def handleUnsubscribe(self, ffrom):
        """This function is called by the IMCom library when someone
        is requesting to remove us from their roster."""
        self.output(usercolor + ffrom + defaultcolor + " is unsubscribing from your presence.")

    # Reports whether a change to the roster was successful or not
    def handleRosterUpdateCheck(self, successful, message):
        """This function is called by the IMCom library after a roster
        change has been made, telling us whether or not it was successful"""
        if(successful):
            self.output("Roster update successful: " + message)
        else:
            self.output("Roster update " + errorcolor + " Error: " +
                        message + defaultcolor)

    # Called when there has been a change to the roster
    def handleRosterUpdate(self, jid, name, subs, ask, success):
        """This function is called by the IMCom library when another
        resource has made a change to our roster.
        """
        # update our roster lists based upon information in jid, name
        if(subs == "remove"):
            if(success):
                self.output("Removed " + usercolor + jid + defaultcolor +
                            " from our lists")
                return
            else:
                self.output("Removal process " + errorcolor +
                            "failed!" + defaultcolor)
                return

    def handleNickCollision(self, nickname, newjid, oldjid):
      self.output(errorcolor + "WARNING:" + defaultcolor + " A nickname collision occured.\n" +
                  "Nickname: " + keycolor + nickname + defaultcolor + " is now linked to JID: " +
                  keycolor + newjid + defaultcolor + " instead of " + keycolor + oldjid + defaultcolor)

    def handleVCardSubmit(self, success):
        if success:
            self.output("Successfully submitted your vcard")
        else:
            self.output("Your VCard submission was unsuccessful.")

    def handleVCard(self, ffrom, fn, given, family, nickname, email):
        self.appendUserQueue(ffrom)
        if(ffrom == None):
            return
        self.output("Information on " + usercolor + ffrom +
                    defaultcolor + ": ")
        if(fn != None):
            self.output(keycolor + "Fullname: " + desccolor + fn +
                        defaultcolor)
        if(given != None):
            self.output(keycolor + "Given: " + desccolor + given +
                        defaultcolor)
        if(family != None):
            self.output(keycolor + "Family: " + desccolor +
                        family + defaultcolor)
        if(nickname != None):
            self.output(keycolor + "Nickname: " + desccolor +
                        nickname + defaultcolor)
        if(email != None):
            self.output(keycolor + "Email: " + desccolor +
                        email + defaultcolor)

    def handleNoVCard(self, ffrom):
        self.appendUserQueue(ffrom)
        if(ffrom == None):
            return
        self.output("There is no information (no vcard filed) for " +
                    usercolor + ffrom + defaultcolor + ". ")
        return

    def handleNegotiationRequest(self, ffrom, type, options, id):
      nick = self.getNick(ffrom)
      text = usercolor + nick + defaultcolor + " is requesting (id: " + keycolor + id + defaultcolor + ") "
      text = text + "which of the following works for you for " + desccolor + type + defaultcolor + ".\n"
      for option in options:
        text = text + option + "\n"
      self.output(text)

    def handleNegotiationResult(self, ffrom, type, options, id):
      nick = self.getNick(ffrom)
      text = usercolor + nick + defaultcolor + " has specified (id: " + keycolor + id + defaultcolor + ") "
      text = text + "that " + keycolor + options[0] + defaultcolor + " works for them for " + desccolor + type + defaultcolor
      self.output(text)


    def handleVersionResponse(self, ffrom, fromResource, name, version, os):
      nick = self.getNick(ffrom)
      text = usercolor + nick + "/" + fromResource + defaultcolor + " is using client: " + desccolor + name + defaultcolor
      if version != None:
        text = text + " version: " + keycolor + version + defaultcolor
      if os != None:
        text = text + " on os: " + keycolor + os + defaultcolor
      self.output(text)

    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # -------------------------------------------------------------------------
    #  Utility Functions
    # -------------------------------------------------------------------------
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++



    def addSocketWrapper(self, sw):
        self.socketlisteners.append(sw)
        self.output("A client socket started.")

    def removeSocketWrapper(self, sw):
        self.socketlisteners.remove(sw)
        self.output("A client socket disconnected.")

    def getColor(self, fg, bg):
        retval = fg[:-1] + ";" + bg[4:]
        return retval


    def setColors(self):
        global usercolor, statuscolor, errorcolor, messagebodycolor,\
               timecolor, desccolor, sepcolor, defaultcolor, keycolor,backgroundcolor

        g = self.profile.sessioncolors

        backgroundcolor = g["background"]

        usercolor = self.getColor(g["user"], backgroundcolor)
        statuscolor = self.getColor(g["status"], backgroundcolor)
        errorcolor = self.getColor(g["error"], backgroundcolor)
        messagebodycolor = self.getColor(g["messagebody"], backgroundcolor)
        timecolor = self.getColor(g["time"], backgroundcolor)
        desccolor = self.getColor(g["desc"], backgroundcolor)
        sepcolor = self.getColor(g["sep"], backgroundcolor)
        defaultcolor = self.getColor(g["default"], backgroundcolor)
        keycolor = self.getColor(g["key"], backgroundcolor)

        self.profile.switches['colors'][0] = "true"
        setCIColors(self.prefs, self.profile)

    def setNoColors(self):
        global usercolor, statuscolor, errorcolor, messagebodycolor,\
               timecolor, desccolor, sepcolor, defaultcolor, keycolor, backgroundcolor
        usercolor = ""
        statuscolor = ""
        errorcolor = ""
        messagebodycolor = ""
        timecolor = ""
        desccolor = ""
        sepcolor = ""
        defaultcolor = ""
        keycolor = ""
        backgroundcolor = ""
        self.profile.switches['colors'][0] = "false"


    def beginLine(self):
      if(self.profile.switches['clearline'][0] == "true"):
        output=chr(27)+"[%dD"%80 + backgroundcolor
        self.printText(output)

    def clearCommand(self):
      self.beginLine()
      if(self.profile.switches['clearline'][0] == "true"):
        output=chr(27)+"[K" + backgroundcolor
        self.printText(output)
      else:
        self.printText("\n")

    def clearScreen(self):
        if(self.profile.switches['colors'][0] == "true"):
            output = defaultcolor + chr(27)+"[J" + defaultcolor
            sys.stdout.write(output)
            sys.stdout.flush()

    def statusNag(self):
        if( self.currentStatus != "online" ):
            self.output( "Your current status is " +
                statuscolor + self.currentStatus + defaultcolor +
                ": " + desccolor +
                self.currentStatusReason + defaultcolor )

    def printText(self, st):
        try:
            sys.stdout.write(st.encode(self.profile.encoding, 'replace'))
            sys.stdout.flush()
        except:
            traceback.print_exc()
            print "An exception occuring trying to print to terminal."
            print "This is most likely due to an extended ascii character."
            print "You may have missed a message or query response."

    def dumpQueue(self):
        for item in self.outputQueue:
            self.clearCommand()
            self.printText(item+"\n")
        self.outputQueue = []


    def output(self, st):
        st = defaultcolor + st
        for key in self.socketlisteners:
            key.sendData(st+"\n.\n")
        if(( self.mode == self.MULTILINE) and (  not self.allowInterrupt )):
            self.outputQueue.append(st)
            return
        self.clearCommand()
        self.printText(st+"\n")
        if(READLINE == 1 and self.gettingCommand):
            self.printText(self.prompt + readline.get_line_buffer().decode(self.profile.encoding))
        elif(self.gettingCommand):
            self.printText(self.prompt)


    def checkto(self, nick):
        if(-1 != string.find(nick,"@")):
            return 1
        if self.imcom.nickhash.has_key(nick):
            return 1
        if self.imcom.confnick.has_key(nick):
            return 1
        tmp = string.split(nick,".")
        if(len(tmp) > 1):
            return 1
        return 0

    def getTime(self):
        return time.strftime('%H:%M:%S',time.localtime(time.time()))

    def getDate(self):
        return time.strftime('%m/%d/%Y',time.localtime(time.time()))

    # begin add by ted
    def logMultiMessage(self, fnames, sender, ddate, ttime, text):
        for fname in fnames:
            self.logMessage(fname, sender, ddate, ttime, text)
    # end add by ted

    def logMessage(self, fname, sender, ddate, ttime, text):
        path = os.environ['HOME']+"/.imcom/"+self.profile.name+"/"

        if( sender == self.profile.user+"@"+self.profile.server ):
            self.lastToDict[fname] = text
        else:
            self.lastFromDict[sender] = text

        # strip filename of /'s from resource string..
        if(-1 != fname.find("/")):
            fname = fname[:fname.find("/")]

        # fix the text by replacing < with &lt; etc.

        text = self.imcom.normalize(text)

        # make sure path exists.
        try:
            os.makedirs(path[:-1],0700)
        except:
            pass
        try:
            if (fname != self.getNick(fname)):
                os.symlink(fname,path+self.getNick(fname))
        except:
            pass
        try:
            if(CODECS):
                f = codecs.open(path+fname,"ab+", self.profile.encoding, 'replace')
            else:
                f = open(path+fname,"ab+")
            f.write("<message from='"+sender+"' date='"+ddate+
                    "' time='"+ttime+"'>")
            f.write(text)
            f.write("</message>\n")
            f.flush()
            f.close()
        except:
            print "There was an exception attempting to log the message"
            traceback.print_exc()


    def getJID(self, nick):
        result = nick
        resource = None
        if nick == None or len(nick) == 0:
          return (None, None)
        if(nick != None and -1 != string.find(nick,"/")):
            resource = nick[string.find(nick,'/')+1:]
            nick = nick[:string.find(nick,'/')]
        else:
            resource = None
        if(nick[0] == "$" and len(nick) == 2):
            n = int(nick[1])
            if(len(self.userQueue)>n):
                result = self.userQueue[n]
                if(self.debug):
                    logDebug("$? Result is: " + result)
                return (result, self.imcom.normalize(resource))
            else:
                result = nick
        if(self.imcom.nickhash.has_key(nick)):
            result = self.imcom.nickhash[nick]
        elif(self.imcom.confnick.has_key(nick)):
            result = self.imcom.confnick[nick]
        else:
            result = nick
        if(self.imcom.jidhash.has_key(result) and self.imcom.reshash.has_key(result) and resource == None):
          resource = self.imcom.reshash[result][0]
        return (result, self.imcom.normalize(resource))

    def isAgent(self, jid):
        return self.imcom.agenthash.has_key(jid)

    def getNick(self, jid):
        result = jid
        if jid.find("/") != -1:
          jid = jid[:jid.find("/")]
        if( string.find( jid, "@" ) == -1 ):
            if( self.imcom.agenthash.has_key(jid) ):
                return self.imcom.agenthash[jid]
        if(self.imcom.jidhash.has_key(jid) and
           self.imcom.jidhash[jid] != None):
            result = self.imcom.jidhash[jid]
        else:
            result = jid
        return result

    def updateKeyInHash(self, hash, oldkey, newkey):
        keys = hash.keys()
        for item in keys:
            if(item == oldkey):
                value = hash[item]
                hash[newkey]=value
                return

    def updateHash(self, hash, oldkey, newkey, value):
        keys = hash.keys()
        for item in keys:
            if(item == oldkey):
                del hash[item]
        hash[newkey] = value


    def appendUserQueue(self, ffrom):
        self.userQueue.insert(0,ffrom)
        if(len(self.userQueue) > 10):
            self.userQueue = self.userQueue[-10:]
        return


    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # -------------------------------------------------------------------------
    #  XData Functions
    # -------------------------------------------------------------------------
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    def createXDataLabel(self, var, label, required, default):
      myLabel = desccolor

      if label != None:
        myLabel = myLabel + label
      else:
        myLabel = myLabel + var

      if default != None:
        myLabel = myLabel + defaultcolor + " [" + keycolor + str(default) + defaultcolor + "] : "
      else:
        myLabel = myLabel + defaultcolor + " : "

      return myLabel


    def getXDataListSingleAnswer(self, var, label, default, options, required):
      lstlen = len(options)
      print "You may answer with a number between 1 and", lstlen
      myLabel = self.createXDataLabel(var, label, required, None)

      n = 1
      for option in options:
        print keycolor, n, defaultcolor + ":" + desccolor, option[0], defaultcolor
        n = n + 1

      result = None
      print myLabel
      result = raw_input( "> " )

      if result == '' and not required:
        return [var, '']

      answer = None
      try:
        answer = int(result)
      except:
        pass

      while answer == None or answer > lstlen or answer < 1:
        print "Please answer with a number between 1 and", lstlen
        print myLabel
        result = raw_input("> " )
        if result == '' and not required:
          return [var, '']

        answer = None
        try:
          answer = int(result)
        except:
          pass

      return [var, options[answer-1][1]]


    def getXDataBooleanAnswer(self, var, label, value, required):
      print "You may answer with (yes, true, 1, y, t), (no, false, 0, n, f)\nJust hit enter if you don't want to answer at all."
      myLabel = self.createXDataLabel(var, label, required, value)
      result = None
      answer = None

      print myLabel
      result = raw_input( "> " )
      if value != None and result == '':
        result = value

      if result.lower() in ("yes", "true", "1", "y", "t"):
        answer = 1
      if result.lower() in ("no", "false", "0", "n", "f"):
        answer = 0

      if result == '':
        answer = ""

      while (required and result == None) or answer == None:
        if required:
          print "This is a required field, please enter a valid answer."
        else:
          print "Please answer with (yes, true, 1, y, t), (no, false, 0, n, f)\nJust hit enter if you don't want to answer at all."
        print myLabel
        result = raw_input( "> " )
        if result.lower() in ("yes", "true", "1", "y", "t"):
          answer = 1
        if result.lower() in ("no", "false", "0", "n", "f"):
          answer = 0
        if result == '':
          answer = ""

      return [var, str(answer)]


    def getXDataTextSingleAnswer(self, var, label, value, required, private):
      myLabel = self.createXDataLabel(var, label, required, value)
      result = None

      print myLabel
      if private:
        result = getpass.getpass( "> " )
      else:
        result = raw_input( "> " )

      if value != None and result == '':
        result = value

      while required and result == '':
        print "This is a required field, please enter it correctly."
        print myLabel
        if private:
          result = getpass.getpass( "> " )
        else:
          result = raw_input( "> " )

      return [var, result]


    def getXDataTextMultiAnswer(self, var, label, value, required):
      myLabel = self.createXDataLabel(var, label, required, value)
      result = ""
      done = None
      print "Please enter information, finishing with a single period on a line of its own."
      print myLabel
      while(not done):
        tmp = raw_input( "> " )
        if tmp == ".":
          done = 1
          continue
        result = result + "\n" + tmp
      return [var, result]


    def getXDataListMultiAnswer(self, var, label, default, options, required):
      myLabel = self.createXDataLabel(var, label, required, None)
      result = ''
      lstlen = len(options)

      print "Please select as many options as you want 1-"+str(lstlen)+" delimiting by commas only, no spaces."

      n = 1
      for option in options:
        print keycolor, n, defaultcolor + ":" + desccolor, option[0], defaultcolor
        n = n + 1

      print myLabel
      tmp = raw_input( "> " )
      tmps = tmp.split(",")
      for num in tmp:
        try:
          i = int(num)
          if i < 1 or i > lstlen:
            continue
          result = result + "\n" + options[i][1]
        except:
          pass
      return result


    def getXDataJIDSingleAnswer(self, var, label, value, required):
      myLabel = self.createXDataLabel(var, label, required, value)
      result = None
      gaveGarbage = 0

      print "You may either use a fully qualified JID, or the nick name of someone on your roster."

      print myLabel
      result = raw_input( "> " )

      if value != None and result == '':
        result = value

      # validate jid here by nick or jid validation routine
      jid = self.getJID(result)[0]
      if(not self.checkTo(jid)):
        gaveGarbage = 1

      while (required and result == '') or gaveGarbage:
        print "This is a required field, please enter it correctly."
        print myLabel
        result = raw_input( "> " )
        # validate jid here by nick or jid validation routine
        jid = self.getJID(result)[0]
        if(not self.checkTo(jid)):
          gaveGarbage = 1
        else:
          gaveGarbage = 0

      return [var, jid]


    def getXDataJIDMultiAnswer(self, var, label, value, required):
      myLabel = self.createXDataLabel(var, label, required, value)
      result = ''

      print "You may either use a fully qualified JID, or the nick name of someone on your roster."
      print "Separate JIDs with commas only."

      print myLabel
      tmp = raw_input( "> " )

      if value != None and tmp == '':
        result = value

      tmps = tmp.split(',')
      for maybeJid in tmps:
        # validate jid here by nick or jid validation routine
        jid = self.getJID(maybeJid)[0]
        if(self.checkTo(jid)):
          result = result + "\n" + jid

      return [var, result]


    def getXDataAnswers(self, xdataInformation):
      self.output("Please finish whatever command you're currently entering (or just hit enter) to begin filling out this form.")
      self.pendingCommand = 1
      while(self.gettingCommand == 1):
        time.sleep(0.01)
      self.mode = self.MULTILINE
      if(READLINE == 1):
        readline.parse_and_bind("set disable-completion on")
      self.gettingCommand = 1

      answers = []
      for field in xdataInformation:
        print
        required = None
        if field[5] != None:
          required = int(field[5])

        if field[6] != None:
          print desccolor + field[6] + defaultcolor
        if field[0] == "hidden":
          answers.append([field[1], field[3]])
          continue
        if field[0] == "fixed":
          print field[3]
          continue
        if field[0] == "text-single":
          answers.append(self.getXDataTextSingleAnswer(field[1], field[2], field[3], required, 0))
          continue
        if field[0] == "list-single":
          answers.append(self.getXDataListSingleAnswer(field[1], field[2], field[3], field[4], required))
          continue
        if field[0] == "list-multi":
          answers.append(self.getXDataListMultiAnswer(field[1], field[2], field[3], field[4], required))
          continue
        if field[0] == "boolean":
          answers.append(self.getXDataBooleanAnswer(field[1], field[2], field[3], required))
          continue
        if field[0] == "text-private":
          answers.append(self.getXDataTextSingleAnswer(field[1], field[2], field[3], required, 1))
          continue
        if field[0] == "text-multi":
          answers.append(self.getXDataTextMultiAnswer(field[1], field[2], field[3], required))
          continue
        if field[0] == "jid-single":
          answers.append(self.getXDataJIDSingleAnswer(field[1], field[2], field[3], required))
          continue
        if field[0] == "jid-multi":
          answers.append(self.getXDataJIDMultiAnswer(field[1], field[2], field[3], required))
          continue
      self.gettingCommand = 0
      self.pendingCommand = 0
      if(READLINE == 1):
        # turn completions back on
        readline.parse_and_bind("set disable-completion off")
      # show the messages that came in while they were busy
      self.dumpQueue()
      self.mode = 0
      return answers






    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    # -------------------------------------------------------------------------
    #  Command Line Input, Execution Functions
    # -------------------------------------------------------------------------
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


    def run(self):
        self.running = 1
        self.clearScreen()
        printStartup()
        try:
            while(self.running):
                if( len( self.commandQueue ) == 0 ):
                    while self.pendingCommand != 0:
                        self.gettingCommand = 0
                        time.sleep(0.1)
                    self.gettingCommand = 1
                    self.line = raw_input(self.prompt)
                    self.gettingCommand = 0
                else:
                    self.line = self.commandQueue.pop( 0 )
                self.executeCommand(self.line)
        except EOFError:
            print
            self.executeCommand(self.prefs.getCommand(self.profile,"quit"))
        except KeyboardInterrupt:
            print
            self.executeCommand(self.prefs.getCommand(self.profile,"quit"))
            #except:
            #    print "something bad happened"
            #    print "Unexpected error:", sys.exc_info()[0]
            #    self.executeCommand(self.profile.commands.quit)

    def getAllText(self,prompt):
        l = ""
        t = ""
        saveprompt = ""

        self.mode = self.MULTILINE
        if(READLINE == 1):
            readline.parse_and_bind("set disable-completion on")
        self.gettingCommand = 1

        # save the current prompt before we change it
        saveprompt = self.prompt
        self.prompt = "[" + prompt + "]: "

        l = raw_input( self.prompt )
        while(l != "."):
            if(l == "#"):
                t = None
                break
            t = t + l + "\n"
            l = raw_input("[" + prompt + "]: ")
        self.gettingCommand = 0
        if(READLINE == 1):
            # turn completions back on
            readline.parse_and_bind("set disable-completion off")
        # show the messages that came in while they were busy
        self.dumpQueue()
        self.mode = 0
        # repair the prompt
        self.prompt = saveprompt
        # remove the last newline?
        if( t != None ):
            t = t[:-1]
            t = t.decode(self.profile.encoding)
        return t

    def checkAlias(self, command, line, profile):
        if(profile.aliases.has_key(command)):
            return profile.aliases[command] + " " + line
        return None



    def executeCommand(self, line):
        p = self.prefs
        pr = self.profile
        if not isinstance(line,unicode):
          line = line.decode(pr.encoding)
        temp = string.split(line,None,1)
        if self.autostatus != None:
          self.autostatus.setIdleTime( 0 )
          self.autostatus.doAutoStatus()
        if(len(temp)==0):
          return
        command = string.lower(temp[0])
        if(len(temp)>1):
            line = temp[1]
        else:
            line = ""
        line = string.strip(line)
        a = self.checkAlias(command, line, pr)
        if(a != None):
            self.executeCommand(a)
            return
        if(command == "/startserver"):
            self.socketserver = SocketWrapper.listener(self)
            self.socketserver.start()
            print "Started ths socket server"
            return
        if(command == "/stopserver"):
            self.socketserver.cli = None
            self.socketserver.running = 0
            print "Stopped the socket server"
            return
        if(command == "/eval"):
            try:
                pp.pprint(eval(line))
            except:
              traceback.print_exc()
              a,b,c = sys.exc_info()
              print "An exception occured"
              print(a)
              print(b)
              print(c)
              print errorcolor + "ERROR: " + defaultcolor + \
                    "This doesn't evaluate well: <" + line + ">"
            return
        keys = pr.sessioncommands.keys()
        for key in keys:
            if command == pr.sessioncommands[key]:
                callstring = key + "Command(self, line)"
                eval(callstring)
                return
        print "What you say?? (" + self.prefs.getCommand(self.profile, "help") +  " for help.)"



    def findItem(self, dict, item):
        keys = dict.keys()
        for blah in keys:
            if(dict[blah] == item):
                return blah
        return None

    def printRoster(self):
        self.printRosterBackend( 0 );

    def printRosterBackend(self,onlineonly):
        trs = ""  # trs is theReturnString
        getTerminalWidth()
        trs = trs + sepcolor + \
              "----------------------------------------" + defaultcolor + "\n"
        trs = trs + "You are currently " + statuscolor + self.imcom.status + \
              defaultcolor + ":" + self.currentStatusReason + "\n"
        trs = trs + sepcolor + \
              "---------------Your Roster--------------" + defaultcolor + "\n"
        keys = self.imcom.jidhash.keys()
        mxn = 0
        mxs = 0
        mxt = 0
        nicks = {}
        for person in keys:
            if((not self.imcom.subscriptions.has_key(person) or
                self.imcom.subscriptions[person] == 'from' or
                self.imcom.subscriptions[person] == 'none') and
               (self.imcom.asks[person] == 'none' or
                self.imcom.asks[person] == None) and
               (not (self.imcom.subscriptions[person] == 'from' and
                     person.find('@') == -1))):
              continue
            if(not self.imcom.gjidhash.has_key(person)):
                continue
            glist = self.imcom.gjidhash[person]
            if(operator.contains(glist,"lurker") or
               operator.contains(glist,"ignore")):
                continue
            nick = self.getNick(person)
            if not self.imcom.preshash.has_key(person):
                continue
            available,show,status = self.imcom.preshash[person]
            if(len(nick) > mxn):
                mxn = len(nick)
            if(len(show) > mxs):
                mxs = len(show)
            if(len(status) > mxt):
                mxt = len(status)
            nicks[nick] = (show, status, available)
        order = ("dnd","xa","away","chat","online")
        keys = nicks.keys()
        if( not onlineonly ):
            for person in keys:
                show,status,a = nicks[person]
                asks = ""
                if(self.imcom.asks.has_key(self.getJID(person)[0]) and
                   self.imcom.asks[self.getJID(person)[0]] == 'subscribe'):
                    asks = " awaiting authorization"
                if(not a):
                    line = usercolor + string.ljust(person,mxn+4) + \
                            defaultcolor + statuscolor + \
                            string.ljust(show,mxs+2) + defaultcolor +\
                            desccolor + status + defaultcolor + asks
                    trs = trs + line + "\n"
        for stat in order:
            for person in keys:
                show,status,a = nicks[person]
                if(a and show == stat):
                    resources = ""
                    asks = ""
                    if(self.imcom.asks.has_key(self.getJID(person)[0]) and
                       self.imcom.asks[self.getJID(person)[0]] == 'subscribe'):
                        asks = " awaiting authorization"
                    if(self.imcom.reshash.has_key(self.getJID(person)[0])):
                        for res in self.imcom.reshash[self.imcom.nickhash[person]][2].keys():
                            ress = self.imcom.reshash[self.imcom.nickhash[person]][2][res][1][0]
                            if(resources == "" and res != ""):
                                resources = " from resources: (" + keycolor + res + \
                                            defaultcolor + ", " + statuscolor + ress + \
                                            defaultcolor + ")"
                            elif(resources != ""):
                                resources = resources + ", (" + keycolor + res + \
                                            defaultcolor + ", " + statuscolor + ress + \
                                            defaultcolor + ")"
                    line = usercolor + string.ljust(person,mxn+4) +\
                           defaultcolor + statuscolor +\
                           string.ljust(show,mxs+2) + defaultcolor +\
                           desccolor + status + defaultcolor + resources + asks
                    trs = trs + line + "\n"
        self.output(trs)

    def printOnlineRoster(self):
        self.printRosterBackend( 1 );

    def printOnlineRosterOldDelete(self):
        getTerminalWidth()
        self.output(sepcolor +
              "----------------------------------------" + defaultcolor)
        self.output("You are currently " + statuscolor + self.imcom.status +
              defaultcolor + " : " + self.currentStatusReason)
        self.output(sepcolor +
              "--------------Users Online--------------" + defaultcolor)
        keys = self.imcom.jidhash.keys()
        mxn = 0
        mxs = 0
        mxt = 0
        nicks = {}
        for person in keys:
            if(not self.imcom.gjidhash.has_key(person)):
                continue
            glist = self.imcom.gjidhash[person]
            if(operator.contains(glist,"lurker") or
               operator.contains(glist,"ignore")):
                continue
            nick = self.getNick(person)
            if(not nick):
                continue
            available,show,status = self.imcom.preshash[person]
            if(len(nick) > mxn):
                mxn = len(nick)
            if(len(show) > mxs):
                mxs = len(show)
            if(len(status) > mxt):
                mxt = len(status)
            nicks[nick] = (show, status, available)
        order = ("dnd","xa","away","chat","online")
        keys = nicks.keys()
        for stat in order:
            for person in keys:
                show,status,a = nicks[person]
                if(a and show == stat):
                    resources = ""
                    if(self.imcom.reshash.has_key(self.imcom.nickhash[person])):
                        for res in self.imcom.reshash[self.imcom.nickhash[person]][2].keys():
                            ress = self.imcom.reshash[self.imcom.nickhash[person]][2][res][1][0]
                            if(resources == "" and res != ""):
                                resources = " from resources: (" + keycolor + res + \
                                            defaultcolor + ", " + statuscolor + ress + \
                                            defaultcolor + ")"
                            elif(resources != ""):
                                resources = resources + ", (" + keycolor + res + \
                                            defaultcolor + ", " + statuscolor + ress + \
                                            defaultcolor + ")"
                    line = usercolor + string.ljust(person,mxn+4) +\
                           defaultcolor + statuscolor + \
                           string.ljust(show,mxs+2) + defaultcolor +\
                           desccolor + status + defaultcolor + resources
                    self.output(line)

    def showGroupMembers(self, groupname):
        self.printText(sepcolor + "----------------------------------------" + defaultcolor + "\n")
        self.printText(keycolor + string.center(groupname,40) + defaultcolor + "\n")
        self.printText(sepcolor + "----------------------------------------" + defaultcolor + "\n")
        keys = self.imcom.grouphash[groupname]
        mxn = 0
        mxs = 0
        mxt = 0
        nicks = {}
        for person in keys:
            glist = self.imcom.gjidhash[person]
            if(groupname != "lurker" and
               operator.contains(glist,"lurker")):
                continue
            if(groupname != "ignore" and
               operator.contains(glist,"ignore")):
                continue
            #if(self.imcom.subscriptions[person] == "from" or
            #   self.imcom.subscriptions[person] == "none" or
            #   self.imcom.subscriptions[person] == None):
            #    continue
            nick = self.getNick(person)
            available,show,status = self.imcom.preshash[person]
            if(len(nick) > mxn):
                mxn = len(nick)
            if(len(show) > mxs):
                mxs = len(show)
            if(len(status) > mxt):
                mxt = len(status)
            nicks[nick] = (show, status, available)
        order = ("dnd","xa","away","chat","online")
        keys = nicks.keys()
        for person in keys:
            show,status,a = nicks[person]
            if(not a):
                line = usercolor + string.ljust(person,mxn+4) + defaultcolor +\
                       statuscolor + string.ljust(show,mxs+2) + defaultcolor +\
                       desccolor + string.ljust(status,mxt) + defaultcolor
                self.printText(line+"\n")
        for stat in order:
            for person in keys:
                show,status,a = nicks[person]
                if(a and show == stat):
                    resources = ""
                    if(self.imcom.reshash.has_key(self.getJID(person)[0])):
                        for res in self.imcom.reshash[self.getJID(person)[0]][2].keys():
                            ress = self.imcom.reshash[self.getJID(person)[0]][2][res][1][0]
                            if(resources == "" and res != ""):
                                resources = " from resources: (" + keycolor + \
                                            res + defaultcolor + ", " + \
                                            statuscolor + ress + \
                                            defaultcolor + ")"
                            elif(resources != ""):
                                resources = resources + ", (" + keycolor + \
                                            res + defaultcolor + ", " + \
                                            statuscolor + ress + \
                                            defaultcolor + ")"
                    line = usercolor + string.ljust(person,mxn+4) +\
                           defaultcolor + statuscolor +\
                           string.ljust(show,mxs+2) + defaultcolor +\
                           desccolor + status + defaultcolor + resources
                    self.printText(line + "\n")

    def timeInRange( self, unixTime, beginTime, endTime ):
        curtime = time.localtime( unixTime )
        beginTime = 0, 0, 0, int( beginTime[0] ), int( beginTime[1] ), \
            0, 0, 0, curtime[8]
        endTime = 0, 0, 0, int( endTime[0] ), int( endTime[1] ), \
            0, 0, 0, curtime[8]
        curtime = 0, 0, 0, curtime[3], curtime[4], 0, 0, 0, curtime[8]

        if( ( time.mktime( curtime ) >= time.mktime( beginTime ) ) and
            ( time.mktime( curtime ) <= time.mktime( endTime ) ) ):
            return 1

        return 0

    def doAutoStatus( self, idleTime ):
        self.output( "Wow, michael just made a serious error. Please "\
            "message him at michael@www.cosby.dhs.org to "\
            "mention this." )
        return

def tabCompleter(text, state):
    global cli
    try:
        if(text == None):
            return None
        params = string.split(readline.get_line_buffer())
        if(len(params) == 1 and readline.get_line_buffer()[-1] != ' '):
            # We're tab completing a command
            r = completeFromHash(cli.profile.aliases,text,state)
            if(r == None):
                return completeFromHashValue(cli.profile.sessioncommands,
                                             text,state)
            else:
                return r
        if((len(params) == 1 and readline.get_line_buffer()[-1] == ' ') or
           (len(params) == 2 and readline.get_line_buffer()[-1] != ' ')):
            # We're probably completing a nick
            return completeFromHash(cli.imcom.nickhash,text,state)
        if((len(params) == 3 and readline.get_line_buffer()[-1] != ' ') or
           (len(params) == 2 and readline.get_line_buffer()[-1] == ' ')):
            # Command specific tab completion
            return None
    except:
        traceback.print_exc()
        a,b,c = sys.exc_info()
        print "An exception occured"
        print(a)
        print(b)
        print(c)

def completeFromHash(hash, text, state):
    keys = hash.keys()
    i=0
    if len(text) > 0 and text[0] in ("'", '"'):
      text = text[1:]
    for item in keys:
        if (item == None):
            continue
        if (text == item[:len(text)]):
            if(i == state):
                #print "Returning " + item
                if(item.find(' ') != -1):
                  return '"' + item + '"'
                return item
            else:
                i = i + 1
    #print "Returning None"
    return None

def completeFromHashValue(hash, text, state):
    #print "In completeFromHashValue"

    #print "hash is: ", hash
    keys = hash.keys()
    i = 0
    for item in keys:
        val = hash[item]
        if(text == val[:len(text)]):
            if(i == state):
                return val
            else:
                i = i + 1
    return None


def printStartup():
    print("IMCom version " + str(VERSION) + " is released under the BSD license.\nSee LICENSE for details")
    print("Written by Casey Crabb (JID: airog@floobin.cx)")

def logDebug(s):
    #print s.encode(self.profile.encoding, "replace")
    print s.encode('ASCII', "replace")

def printUsage():
  printStartup()
  print("Usage: imcom <name-of-profile-to-use>")
  print("Example: imcom floobin.cx")
  print("Example: imcom")
  print("")

def handler(signum, frame):
    #global cli
    #if cli.debug:
    print 'Signal handler called with signal', signum
    print 'Frame is ', frame
    pass

def quithandler(signum, frame):
    cli.executeCommand(cli.prefs.getCommand(cli.profile,"quit"))

if(__name__ == "__main__"):
    global cli
    # if LOCALE:
    #     locale.setlocale(locale.LC_ALL, "de") #("US","ISO-8859-1"))
    # signal.signal(signal.SIGWINCH,handler)
    # signal.signal(signal.SIGHUP,handler)
    signal.signal(signal.SIGINT,quithandler)
    # signal.signal(signal.SIGPIPE,handler)
    # signal.signal(signal.SIGUSR1,handler)
    # signal.signal(signal.SIGUSR2,handler)
    # signal.signal(signal.SIGCHLD,handler)
    # signal.signal(signal.SIGCONT,handler)
    # signal.signal(signal.SIGTRAP,handler)
    # signal.signal(signal.SIGIOT,handler)
    getTerminalWidth()
    # printStartup()

    # remove the old debug log file
    try:
        os.remove(os.environ['HOME']+"/.imcom/debuglog.log")
    except:
        pass

    cli = CLI()
    if(len(sys.argv) > 1):
      if(sys.argv[1] in ("--help", "-h")):
        printUsage()
        sys.exit(0)
      else:
        profile = cli.prefs.getProfile(sys.argv[1])
        if(profile != None):
            cli.profile = profile

    result = cli.applyProfileWrapper()
    if result[0] != 0:
      sys.exit(-1)
    cli.applyPrefs()
    cli.start()

def yesnostr( val ):
    if( val ):
        return "yes"
    else:
        return "no"

def truefalsestr( val ):
    if( val ):
        return "true"
    else:
        return "false"

