import os, sys, socket, time, traceback
import random
from threading import *

debug = 0

def recvTillLF(theSock):
    mystring = ""
    while len(mystring) == 0 or mystring[-1] != '\n':
      tmp = theSock.recv(1)
      if(len(tmp) == 0):
        return mystring
      mystring = mystring + tmp
    return mystring[:-1]

def recvTillEnd(theSock):
  theData = ""
  tmpData = theSock.recv(1024)
  tmp = len(tmpData)
  while tmp != 0:
    theData = theData + tmpData
    tmpData = theSock.recv(1024)
    tmp = len(tmpData)
  return theData

def logDebug(s):
  if debug:
    if isinstance(s, unicode):
        print s.encode(sys.getdefaultencoding(),'replace')
        sys.stdout.flush()
    else:
        print s
        sys.stdout.flush()

class DTCPSocketManager(Thread):
  def __init__(self, host, listenPort):
    Thread.__init__(self)
    self.listenPort = listenPort
    self.host = host
    self.dtcpKeys = {} # keys are the DTCPKey we're looking for upon connection, values are (DTCPKey, bWeIQSET, actionObject)
    self.listeningKeys = [] # our dtcpkeys that we should be listening for
    self.sendingKeys = [] # our dtcpKeys that we are sending.
    self.setDaemon(1)
    self.shouldListen = 1

  def addKey(self, ourDTCPKey, bWeIQSet, theirDTCPKey, actionObject):
    self.dtcpKeys[ourDTCPKey] = [theirDTCPKey, bWeIQSet, actionObject]
    self.listeningKeys.append(ourDTCPKey)
    self.sendingKeys.append(ourDTCPKey)

  def removeKey(self, ourDTCPKey):
    if self.dtcpKeys.has_key(ourDTCPKey):
      del self.dtcpKeys[ourDTCPKey]
    if ourDTCPKey in self.listeningKeys:
      self.listeningKeys.remove(ourDTCPKey)
    if ourDTCPKey in self.sendingKeys:
      self.sendingKeys.remove(ourDTCPKey)

  def removeListeningKey(self, ourDTCPKey):
    if ourDTCPKey in self.listeningKeys:
      self.listeningKeys.remove(ourDTCPKey)
    if not ourDTCPKey in self.sendingKeys:
      self.removeKey(ourDTCPKey)

  def removeSendingKey(self, ourDTCPKey):
    if ourDTCPKey in self.sendingKeys:
      self.sendingKeys.remove(ourDTCPKey)
    if not ourDTCPKey in self.listeningKeys:
      self.removeKey(ourDTCPKey)

  def run(self):
    self.listenSocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.listenSocket.bind((self.host, self.listenPort))
    self.listenSocket.listen(5)
    while 1:
      (theSock, theAddress) = self.listenSocket.accept()
      logDebug("Handling a connection from: " + str(theAddress))
      self.handleConnection(theSock)


  def handleConnection(self, theSock):
    tmp = DTCPSocketListener(self, theSock)
    tmp.start()


  def createConnection(self, ffrom, hosts, bWeIQSet, theirKey, ourKey, actionObject, IMCom):
    tsc = DTCPSocketCreator(self, ffrom, hosts, bWeIQSet, theirKey, ourKey, actionObject, IMCom)
    tsc.start()


class DTCPSocketCreator(Thread):
  def __init__(self, manager, ffrom, hosts, bWeIQSet, theirKey, ourKey, actionObject, IMCom):
    Thread.__init__(self)
    self.manager = manager
    self.ffrom = ffrom
    self.hosts = hosts
    self.bWeIQSet = bWeIQSet
    self.theirKey = theirKey
    self.ourKey = ourKey
    self.actionObject = actionObject
    self.imcom = IMCom
    self.setDaemon(1)

  def run(self):
    for ahost in self.hosts:
      theSock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      try:
        tmp = ahost[0].find(":")
        theHost = ahost[0][:tmp]
        thePort = int(ahost[0][tmp+1:])
        logDebug("sc: attempting to connect to (" + theHost + ", " + str(thePort) + ")")
        theSock.connect((theHost, thePort))
        logDebug("sc: sending theirkey")
        theSock.send("key:"+self.theirKey+"\n")
        command = recvTillLF(theSock)

        logDebug("sc: command is: " + command)

        if command == "error":
          continue

        if command[:3] == "ok:":
          if command[3:] == self.ourKey:
            logDebug("sc: our key matched")
            if self.bWeIQSet:
              logDebug("sc: we have to send the final ok.")
              theSock.send("ok\n")
            if self.manager.dtcpKeys.has_key(self.ourKey):
              self.manager.removeKey(self.ourKey)
              logDebug("sc: all done, sending actionObject on its way")
              self.actionObject.setStreamSocket(theSock)
              return 1
            else:
              logDebug("sc: couldn't finish because we were handled elsewhere.")
              self.manager.removeSendingKey(self.ourKey)
              theSock.close()
              return 0
        theSock.close()
      except:
        traceback.print_exc()
        continue
    self.manager.removeSendingKey(self.ourKey)
    if(len(self.hosts) > 0):
      self.imcom.sendDTCPConnectError(self.ffrom, self.theirKey)

class DTCPSocketListener(Thread):
  def __init__(self, manager, theSock):
    Thread.__init__(self)
    self.manager = manager
    self.theSock = theSock

  def run(self):
    done = 0
    while not done:
      try:
        command = recvTillLF(self.theSock)
      except:
        return

      logDebug("sl: recved command: " + command)
      # tls
      if command == "starttls":
        self.theSock.send("error\n")
        continue

      # initial key
      if command[:4] == "key:":
        if not self.manager.dtcpKeys.has_key(command[4:]):
          logDebug("sl: key was not in our list of keys, close this socket.")
          self.theSock.close()
          return

        if self.manager.dtcpKeys.has_key(command[4:]):
          logDebug("sl: found a matching key")
          ourDTCPKey = command[4:]
          while self.manager.dtcpKeys.has_key(ourDTCPKey) and self.manager.dtcpKeys[ourDTCPKey][0] == None:
            logDebug("sl: waiting on recieving their key: loop c")
            time.sleep(0.10)
          if not self.manager.dtcpKeys.has_key(ourDTCPKey):
            logDebug("sl: our key disappeared, closing socket.")
            self.theSock.close()
            return

          logDebug("sl: sending their key")
          self.theSock.send("ok:"+str(self.manager.dtcpKeys[ourDTCPKey][0]) + "\n")
          if not self.manager.dtcpKeys[ourDTCPKey][1]:
            logDebug("sl: we have to wait for the final ok.")
            command = recvTillLF(self.theSock)
            if command != "ok":
              logDebug("sl: final wasn't 'ok', closing Socket")
              self.theSock.close()
              return
          logDebug("sl: all done, setting action object on its way")
          actionListener = self.manager.dtcpKeys[ourDTCPKey][2]
          if self.manager.dtcpKeys.has_key(ourDTCPKey):
            self.manager.removeKey(ourDTCPKey)
            actionListener.setStreamSocket(self.theSock)
          return
        else:
          logDebug("sl: no matching key")
          self.theSock.close()
          return
      else:
        continue

class DTCPSocketEchoer(Thread):
  def __init__(self, bConnector):
    Thread.__init__(self)
    self.bConnector = bConnector

  def setStreamSocket(self, theSock):
    self.theSock = theSock
    logDebug("set StreamSocket")
    self.start()

  def run(self):
    try:
      if self.bConnector:
        self.theSock.send("Repeater beginning\n")
        command = recvTillLF(self.theSock)
        logDebug("repeater recved: " + command)
      else:
        command = recvTillLF(self.theSock)
        logDebug("repeater recved: " + command)
        self.theSock.send("I'm a repeater: " + command + "\n")

    except:
      pass
    self.theSock.close()
    logDebug("repeater closed socket")


class DTCPSocketBogusXFer(Thread):
  def __init__(self, bConnector):
    Thread.__init__(self)
    self.bConnector = bConnector

  def setStreamSocket(self, theSock):
    self.theSock = theSock
    logDebug("set StreamSocket DTCPSocketBogusXFer: bConnector: " + str(self.bConnector))
    self.start()

  def run(self):
    if self.bConnector:
      logDebug("Generating 32k of bogus data")
      data = ""
      while len(data) != 32768:
        data = data + str(int(random.random() * 1000) % 10)
      logDebug("Sending 32k of bogus data")
      self.theSock.send(data)
      self.theSock.close()
      logDebug("Good to go!")
      return
    else:
      logDebug("expecting 32k of data")
      data = recvTillEnd(self.theSock)
      logDebug("recv()'ed " + str(len(data)) + " bytes of data")
      self.theSock.close()
      logDebug("Good to go!")
      return
