import GUIbackend
from wxPython.wx import *
import time

eventsQueueHash = {}
imcom = None
theListJidDict = {}

IM_RECEIVEEVENT = 10
IM_SENDMESSAGE = 11
IM_GETINFO = 12
IM_REMOVEUSER = 13
IM_ADDUSER = 14

def getNickOrNone(jid):
    if imcom.jidhash.has_key(jid):
        return imcom.jidhash[jid]
    return None

def getNick(jid):
    if imcom.jidhash.has_key(jid):
        return imcom.jidhash[jid]
    return jid

def getTime():
    return time.strftime('%H:%M:%S',time.localtime(time.time()))

def getDate():
    return time.strftime('%m/%d/%Y',time.localtime(time.time()))

class GUITree(wxTreeCtrl):
    def __init__(self, parent, ID, name):
        wxTreeCtrl.__init__(self, parent, ID)
        self.name = name
        self.root = self.AddRoot(name)
        self.SetPyData(self.root, None)
        self.Expand(self.root)

    def addUser(self, jid):
        print "addUser: "
        print "  jid  : ", jid
        nick = getNick(jid)
        print "  nick : ", nick
        child = self.AppendItem(self.root, nick)
        self.SetPyData(child, jid)
        self.setColorForItem(child)
        self.Expand(self.root)

    def removeUser(self, jid):
        item = self.findItemByJid(jid)
        if(item != None):
            self.Delete(item)

    def getJidByPoint(self, pt):
        item, flags = self.HitTest(pt)
        return self.GetItemData(item).GetData()

    def getJidByItem(self, item):
        return self.GetItemData(item).GetData()

    def selectItemByPoint(self, pt):
        item, flags = self.HitTest(pt)
        self.SelectItem(item)

    def setColorForItem(self, item):
        if(item != None):
            jid = self.GetPyData(item)
            if(eventsQueueHash.has_key(jid) and
               len(eventsQueueHash[jid]) != 0):
                print " setting color to red by eventsQueueHash"
                self.SetItemTextColour(item,wxRED)
            elif(imcom.preshash.has_key(jid)):
                print " item         : ", item
                print " preshash[jid]: ", imcom.preshash[jid]
                pres = string.lower(imcom.preshash[jid][1])
                if(pres == 'online' or
                   pres == 'chat'):
                    self.SetItemTextColour(item,wxBLUE)
                elif(pres == 'away'):
                    self.SetItemTextColour(item,wxColor(0,0,127))
                elif(pres == 'xa'):
                    self.SetItemTextColour(item,wxColor(127, 0, 127))
                elif(pres == 'dnd'):
                    self.SetItemTextColour(item,wxColor(64, 0, 64))
                else:
                    self.SetItemTextColour(item, wxBLACK)
            else:
                print "setting color to black "
                self.SetItemTextColour(item,wxBLACK)
            self.SortChildren(self.root)
            self.Refresh()

    def setColorForJid(self, jid):
        print "setColorForJid: "
        print "  jid is: ", jid
        a = self.findItemByJid(jid)
        if(a != None):
            self.setColorForItem(a)


    def doSort(self):
        self.SortChildren(self.root)

    def OnCompareItems(self, item1, item2):
        jid1 = self.GetPyData(item1)
        jid2 = self.GetPyData(item2)
        r = CompareForEvents(jid1,jid2)
        if(r != 0):
            return r
        r = CompareByStatus(jid1, jid2)
        if(r != 0):
            return r
        r = CompareByNick(jid1, jid2)
        return r

    def findItemByJid(self, jid):
        (child, cookie) = self.GetFirstChild(self.root,1)
        while(child.IsOk()):
            print "child is: ", child
            if(self.GetItemData(child).GetData() == jid):
                return child
            (child,cookie) = self.GetNextChild(child, cookie)
        return None

    def updateUser(self, jid, nick):
        item = self.findItemByJid(jid)
        if(item != None):
            self.SetItemText(item, nick)
            #self.setColorForJid(jid)
            self.SortChildren(self.root)
            self.Refresh()

    def hasUser(self, jid):
        item = self.findItemByJid(jid)
        if(item != None):
            return 1
        return 0

    def IsEmpty(self):
        n = self.GetChildrenCount(self.root)
        print "There are", n, "children"
        if(n == 0):
            return 1
        return 0

class SimpleTextFrame(wxFrame):
    def __init__(self, parent, ID, title, text):
        wxFrame.__init__(self, parent, ID, title,
                         wxDefaultPosition, wxSize(350,350))
        panel = wxPanel(self, -1)
        box = wxBoxSizer(wxVERTICAL)
        bbox = wxBoxSizer(wxHORIZONTAL)
        textID = NewId()
        self.textctrl = wxTextCtrl(panel, textID, "")
        box.Add(wxStaticText(panel, -1, text), 0, wxEXPAND | wxALL, 10)
        box.Add(self.textctrl, 1, wxEXPAND | wxALL, 10)
        bbox.Add(wxButton(panel, wxID_OK, " Ok "), 0, wxALL, 10)
        bbox.Add(wxButton(panel, wxID_CANCEL, " Cancel "), 0, wxALL, 10)
        box.Add(bbox, 0, wxEXPAND)
        panel.SetAutoLayout(true)
        panel.SetSizer(box)
        box.Fit(panel)


class UserInfoPanel(wxPanel):
    def __init__(self, parent, ID, jid, nick, ddate, ttime):
        wxPanel.__init__(self, parent, ID)
        # Create the StaticBox which will house all the information
        topInfoComp = wxStaticBox(self, -1, "From:")
        infobox = wxStaticBoxSizer(topInfoComp, wxVERTICAL)
        # Now create the first box that will be put in the StaticBox
        topInfoBox = wxBoxSizer(wxVERTICAL)
        # Add infobox first so we get the StaticBox around our stuff
        topInfoBox.Add(infobox, 1, wxEXPAND)
        # create a horizontal box for position purposes
        infobox2 = wxBoxSizer(wxHORIZONTAL)
        # add the JID information to that box.
        infobox2.Add(wxStaticText(self, -1, "JID: "),
                     0, wxEXPAND|wxALL, 4)
        infobox2.Add(wxStaticText(self, -1, jid,
                                  style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                     2, wxEXPAND|wxALL, 0)

        # if we have a nick, add it to the right of the JID
        if(nick != None):
            infobox2.Add(wxStaticText(self, -1, "Nick Name: "),
                         0, wxEXPAND | wxALL, 4)
            infobox2.Add(wxStaticText(self, -1, nick,
                                      style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                         1, wxEXPAND | wxALL, 0)

        # add the first horizontal layer of information to the Vertical Box
        infobox.Add(infobox2, 1, wxEXPAND|wxALL, 4)

        # create a new horizontal layer
        infobox2 = wxBoxSizer(wxHORIZONTAL)

        # add the date and time to this horizontal layer
        infobox2.Add(wxStaticText(self, -1, "Date: "),
                     0, wxEXPAND|wxALL, 4)
        infobox2.Add(wxStaticText(self, -1, ddate,
                                  style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                     1, wxEXPAND|wxALL, 0)
        infobox2.Add(0,0,1,wxEXPAND)
        infobox2.Add(wxStaticText(self, -1, "Time: "),
                     0, wxEXPAND|wxALL, 4)
        infobox2.Add(wxStaticText(self, -1, ttime,
                                  style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                     1, wxEXPAND|wxALL, 0)

        # add the new horizontal layer to the vertical layer
        infobox.Add(infobox2, 1, wxEXPAND|wxALL,4)

        # Tell the from information panel that it will be layed out
        # automatically by the topInfoBox (which is the StaticBoxSizer)
        self.SetAutoLayout(true)
        self.SetSizer(topInfoBox)
        topInfoBox.Fit(self)

class AddUserFrame(wxFrame): #SimpleTextFrame):
    def __init__(self, parent, ID, defaultNick, defaultJID):
        #SimpleTextFrame.__init__(self,parent,ID, "Add a User",
        #                         "Enter the JID of the user you wish to add to your roster.")
        wxFrame.__init__(self, parent, ID, "Add User",
                         wxDefaultPosition, wxSize(350,350))
        jiddesc = "Enter the JID of the user you wish to add to your roster."
        nickdesc = "Enter the name of the user as you would like it to " + \
                   "appear on your list."
        panel = wxPanel(self, -1)
        box = wxBoxSizer(wxVERTICAL)
        bbox = wxBoxSizer(wxHORIZONTAL)
        textID = NewId()
        self.jidtext = wxTextCtrl(panel, textID, "")
        self.jidtext.SetValue(defaultJID)
        self.nicktext = wxTextCtrl(panel, NewId(), "")
        self.nicktext.SetValue(defaultNick)
        box.Add(wxStaticText(panel, -1, nickdesc), 1, wxALL, 10)
        box.Add(self.nicktext, 1, wxALL, 10)
        box.Add(wxStaticText(panel, -1, jiddesc), 1, wxALL, 10)
        box.Add(self.jidtext, 1, wxALL, 10)
        bbox.Add(wxButton(panel, wxID_OK, " Ok "), 0, wxALL, 10)
        bbox.Add(wxButton(panel, wxID_CANCEL, " Cancel "), 0, wxALL, 10)
        box.Add(bbox, 0, wxEXPAND)
        panel.SetAutoLayout(true)
        panel.SetSizer(box)
        box.Fit(panel)
        EVT_BUTTON(self, wxID_OK, self.OnOK)
        EVT_BUTTON(self, wxID_CANCEL, self.OnCancel)

    def OnOK(self, event):
        imcom.sendSubscribe(self.jidtext.GetValue())
        imcom.setNick(self.jidtext.GetValue(), self.nicktext.GetValue())
        self.Show(false)
        self.Destroy()

    def OnCancel(self, event):
        self.Show(false)
        self.Destroy()


class EventFrame(wxFrame):
    def __init__(self, parent, ID, jid, userEvent):
        wxFrame.__init__(self, parent, ID, "Read a Message",
                         wxDefaultPosition, wxSize(350, 350),
                         style=wxDEFAULT_FRAME_STYLE | wxWANTS_CHARS)

        # Setup Some basic variables
        self.jid = jid
        self.parent = parent
        nick = getNickOrNone(jid)

        # Create the From Information Panel
        fromInfoPanel = UserInfoPanel(self, -1, jid, nick, userEvent[3],
                                      userEvent[4])

        # Create a new box for the messagePanel
        messagePanel = wxPanel(self, -1)
        messageBox = wxBoxSizer(wxVERTICAL)
        textID = NewId()
        # Add the message into the new box
        messageBox.Add(wxTextCtrl(messagePanel, textID, userEvent[1],
                                  style=wxTE_MULTILINE|wxTE_READONLY),
                       1, wxEXPAND | wxALL, 10)
        messagePanel.SetAutoLayout(true)
        messagePanel.SetSizer(messageBox)
        messageBox.Fit(messagePanel)



        # Now create the response panel which will house all the action buttons
        responsePanel = wxPanel(self, -1, style=0)
        responseComp = wxStaticBox(responsePanel, NewId(), "Response Action:")
        responseBox = wxStaticBoxSizer(responseComp, wxHORIZONTAL)

        # Create the buttons
        self.closeButton = wxButton(responsePanel, wxID_CANCEL, " Close ")
        self.replyButton = wxButton(responsePanel, wxID_OK, " Reply ")

        # Add the buttons with a spacer in the middle
        responseBox.Add(self.closeButton, 0, wxALL|wxALIGN_LEFT, 10)
        responseBox.Add(0,0,1,wxEXPAND)
        responseBox.Add(self.replyButton,0, wxALL|wxALIGN_RIGHT, 10)
        responsePanel.SetAutoLayout(true)
        responsePanel.SetSizer(responseBox)
        responseBox.Fit(responsePanel)

        # Create the topBox for holding all 3 panels
        # add the panels in order
        # then set the autolayout.
        topBox = wxBoxSizer(wxVERTICAL)
        topBox.Add(fromInfoPanel,0,wxEXPAND)
        topBox.Add(messagePanel,1,wxEXPAND)
        topBox.Add(responsePanel, 0, wxEXPAND)
        self.SetAutoLayout(true)
        self.SetSizer(topBox)
        topBox.Fit(self)

        EVT_BUTTON(self, wxID_OK, self.OnOK)
        EVT_BUTTON(self, wxID_CANCEL, self.OnCancel)
        EVT_MENU(self, wxID_OK, self.OnOK)
        EVT_MENU(self, wxID_CANCEL, self.OnCancel)
        t = wxACCEL_NORMAL, WXK_ESCAPE, wxID_CANCEL
        responsePanel.SetAcceleratorTable(wxAcceleratorTable([t]))


    def Show(self,b):
        wxFrame.Show(self,b)
        self.SetSize(wxSize(460,325))
        if(b):
            self.replyButton.SetFocus()

    def OnOK(self, event):
        #we're replying to the message...
        self.Show(false)
        self.Destroy()
        self.parent.sendMessageBox(self.jid)

    def OnCancel(self, event):
        self.Show(false)
        self.Destroy()

class MessageFrame(wxFrame):
    def __init__(self, parent, ID, jid):
        wxFrame.__init__(self, parent, ID, "Send a Message",
                         wxDefaultPosition, wxSize(350, 350))
        self.jid = jid
        self.parent = parent
        nick = getNickOrNone(jid)

        toInfoPanel = wxPanel(self, -1)
        toInfoPanel.SetSizeHints(minH=140)
        topInfoComp = wxStaticBox(toInfoPanel, NewId(), "To:")
        topInfoBox = wxBoxSizer(wxVERTICAL)
        infobox = wxStaticBoxSizer(topInfoComp, wxHORIZONTAL)
        topInfoBox.Add(infobox, 1, wxEXPAND)

        infobox.Add(wxStaticText(toInfoPanel, NewId(), "JID: "),
                    0, wxEXPAND | wxALL, 4)
        infobox.Add(wxStaticText(toInfoPanel, NewId(), jid,
                                 style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                    2, wxEXPAND | wxALL, 2)

        if(nick != None):
            infobox.Add(wxStaticText(toInfoPanel, NewId(), "   Nick: "),
                        0, wxEXPAND | wxALL, 4)
            infobox.Add(wxStaticText(toInfoPanel, NewId(), nick,
                                     style=wxSUNKEN_BORDER|wxALIGN_CENTRE),
                        1, wxEXPAND | wxALL, 2)

        infobox.Add(wxButton(toInfoPanel, 12, "Get Info"), 0, wxALL, 1)

        toInfoPanel.SetAutoLayout(true)
        toInfoPanel.SetSizer(topInfoBox)
        topInfoBox.Fit(toInfoPanel)

        panel = wxPanel(self, -1)
        box = wxBoxSizer(wxVERTICAL)
        box2 = wxBoxSizer(wxHORIZONTAL)

        textID = NewId()
        self.textctrl = wxTextCtrl(panel, textID, "", style=wxTE_MULTILINE)
        self.textctrl.SetInsertionPoint(0)
        self.textctrl.SetFocus()
        box.Add(self.textctrl, 1, wxEXPAND, 0)
        box2.Add(wxButton(panel, wxID_OK, " Send "), 0, wxALL, 10)
        box2.Add(wxButton(panel, wxID_CANCEL, " Cancel "), 0, wxALL, 10)
        box.Add(box2, 0, wxALIGN_CENTER)
        panel.SetAutoLayout(true)
        panel.SetSizer(box)
        box.Fit(panel)

        box3 = wxBoxSizer(wxVERTICAL)
        box3.Add(toInfoPanel, 0, wxEXPAND)
        box3.Add(panel, 1, wxEXPAND)

        self.SetAutoLayout(true)
        self.SetSizer(box3)
        box3.Fit(self)
        self.SetSize(wxSize(400,250))
        self.textctrl.SetInsertionPoint(0)
        self.textctrl.SetFocus()
        EVT_BUTTON(self, 12, self.OnGetInfo)
        EVT_BUTTON(self, wxID_OK, self.OnOK)
        EVT_BUTTON(self, wxID_CANCEL, self.OnCancel)
        EVT_MENU(self.textctrl, wxID_OK, self.OnOK)
        EVT_ACTIVATE(self, self.OnActivate)
        EVT_KEY_DOWN(self.textctrl, self.OnKeyDown)
        t = wxACCEL_CTRL, WXK_RETURN, wxID_OK
        s = wxACCEL_ALT, ord("s"), wxID_OK
        self.textctrl.SetAcceleratorTable(wxAcceleratorTable([t,s]))
        self.SetAcceleratorTable(wxAcceleratorTable([t,s]))

    def OnKeyDown(self, event):
        a = event.GetKeyCode()
        print "In onkeydown, a is:", a
        print "ord('s') is: ", ord('s')
        if(a == 83 and
           event.AltDown()):
            self.OnOK(event)
        else:
            event.Skip()

    def OnActivate(self, b):
        if(b.GetActive()):
            self.textctrl.SetFocus()
        b.Skip()

    def OnGetInfo(self, event):
        imcom.sendGetInfo(self.jid)

    def OnOK(self, event):
        #we're sending the message...
        imcom.sendMessage(self.jid, self.textctrl.GetValue(),
                          None)
        self.Show(false)
        self.Destroy()

    def OnCancel(self, event):
        self.Show(false)
        self.Destroy()


def CompareForEvents(jid1, jid2):
    len1 = 0
    len2 = 0
    if(eventsQueueHash.has_key(jid1)):
        len1 = len(eventsQueueHash[jid1])
    if(eventsQueueHash.has_key(jid2)):
        len2 = len(eventsQueueHash[jid2])
    if len1 == len2:
        return 0
    if len1 > 0 and len2 > 0:
        return 0
    if len1 > len2:
        return -1
    return 1

def CompareByStatus(jid1, jid2):
    # chat < online < away < xa < dnd < offline < unavailable
    stat1 = 'unavailable'
    stat2 = 'unavailable'
    if(imcom.preshash.has_key(jid1)):
        stat1 = imcom.preshash[jid1][1]
    if(imcom.preshash.has_key(jid2)):
        stat2 = imcom.preshash[jid2][1]
    stat1 = convertStatusToInt(stat1)
    stat2 = convertStatusToInt(stat2)
    if(stat1 == stat2):
        return 0
    if(stat1 < stat2):
        return -1
    return 1

def convertStatusToInt(stat1):
    if(string.lower(stat1) == "chat"):
        return 0
    if(string.lower(stat1) == "online"):
        return 1
    if(string.lower(stat1) == "away"):
        return 2
    if(string.lower(stat1) == "xa"):
        return 3
    if(string.lower(stat1) == "dnd"):
        return 4
    if(string.lower(stat1) == "offline"):
        return 5
    if(string.lower(stat1) == "unavailable"):
        return 6
    return 7

def CompareByNick(jid1, jid2):
    nick1 = jid1
    nick2 = jid2
    if(imcom.jidhash.has_key(jid1)):
        nick1 = imcom.jidhash[jid1]
    if(imcom.jidhash.has_key(jid2)):
        nick2 = imcom.jidhash[jid2]
    if(nick1 == nick2):
        return 0
    if(nick1 < nick2):
        return -1
    return 1

def CompareListItems(item1, item2):
    jid1 = theListJidDict[item1]
    jid2 = theListJidDict[item2]
    r = CompareForEvents(jid1,jid2)
    if(r != 0):
        return r
    r = CompareByStatus(jid1, jid2)
    if(r != 0):
        return r
    r = CompareByNick(jid1, jid2)
    return r

def findItemInDict(dict, item):
    keys = dict.keys()
    for blah in keys:
        if(dict[blah] == item):
            return blah
    return None





class GUI(wxPanel):
    def __init__(self, parent, name):
        self.parent = parent
        wxPanel.__init__(self, parent, -1, style=wxWANTS_CHARS)
        self.list = None
        self.contextMenu = None
        self.name = name
        tID = NewId()
        self.list = GUITree(self, tID, self.name)
        w, h = self.parent.GetClientSizeTuple()
        self.list.SetDimensions(0,0,w,h)
        EVT_SIZE(self, self.OnSize)
        EVT_LEFT_DCLICK(self.list, self.OnLeftDClick)
        EVT_RIGHT_DOWN(self.list, self.onRightClick)
        EVT_MENU(self, IM_RECEIVEEVENT, self.OnMenuGetNextPending)
        EVT_MENU(self, IM_SENDMESSAGE, self.OnMenuSendMessageBox)
        EVT_MENU(self, IM_GETINFO, self.OnMenuGetInfo)
        EVT_MENU(self, IM_REMOVEUSER, self.OnMenuRemoveUser)
        EVT_MENU(self, IM_ADDUSER, self.OnMenuAddUser)
        EVT_CLOSE(self, self.OnCloseWindow)


    def createList(self):
        # add users to the tree from the roster.
        for item in imcom.jidhash.keys():
            self.list.addUser(item)

        self.list.doSort()
        return

    def finalize(self):
        if(self.contextMenu != None):
            self.contextMenu.Destroy()
            self.contextMenu = None

    def OnCloseWindow(self, event):
        self.finalize()
        event.Skip()

    def OnSize(self, event):
        w, h = self.parent.GetClientSizeTuple()
        self.list.SetDimensions(0,0,w,h)

    def OnLeftDClick(self, event):
        pt = event.GetPosition();
        jid = self.list.getJidByPoint(pt)
        if(jid == None):
            return
        # if we have events pending for this user, display them
        if(eventsQueueHash.has_key(jid) and
           len(eventsQueueHash[jid]) != 0):
            self.getNextPendingEvent(jid)
            return

        # otherwise open up a sendmessage box.
        self.sendMessageBox(jid)


    # This function pops up the Context Menu.
    def onRightClick(self, event):
        jid = self.list.getJidByPoint(event.GetPosition())

        # first select the item we right-clicked...
        self.list.selectItemByPoint(event.GetPosition())
        # create the menu.
        if(self.contextMenu != None):
            self.contextMenu.Destroy()
            self.contextMenu = None
        self.contextMenu = wxMenu("Context Menu", 0)
        if(eventsQueueHash.has_key(jid) and
           len(eventsQueueHash[jid]) > 0):
            self.contextMenu.Append(IM_RECEIVEEVENT,"Receive Event")
        self.contextMenu.Append(IM_SENDMESSAGE, "Send Message")
        self.contextMenu.Append(IM_GETINFO, "Get Info")
        if(self.name == "Others"):
            self.contextMenu.Append(IM_ADDUSER, "Add User")
        else:
            self.contextMenu.Append(IM_REMOVEUSER, "Remove User")

        # then display the menu
        self.PopupMenu(self.contextMenu,
                       event.GetPosition())

    def OnMenuGetNextPending(self, event):
        item = self.list.GetSelection()
        if(item != None):
            jid = self.list.getJidByItem(item)
            self.getNextPendingEvent(jid)

    def OnMenuSendMessageBox(self, event):
        item = self.list.GetSelection()
        if(item != None):
            jid = self.list.getJidByItem(item)
            if(jid != None):
                self.sendMessageBox(jid)

    def OnMenuGetInfo(self, event):
        item = self.list.GetSelection()
        if(item != None):
            jid = self.list.getJidByItem(item)
            if(jid != None):
                imcom.sendGetInfo(jid)

    def OnMenuRemoveUser(self, event):
        item = self.list.GetSelection()
        if(item != None):
            jid = self.list.getJidByItem(item)
            if(jid != None):
                imcom.sendUnsubscribe(jid)

    def OnMenuAddUser(self, event):
        item = self.list.GetSelection()
        if(item != None):
            jid = self.list.getJidByItem(item)
            if(jid != None):
                a = AddUserFrame(self, -1, "", jid)
                a.Show(true)

    def getNextPendingEvent(self, jid):
        # take care of the events pending first.
        userEvent = eventsQueueHash[jid].pop(0)
        self.list.setColorForJid(jid)
        if(userEvent[0] == GUIbackend.IM_MessageReceive):
            EventFrame(self,NewId(),jid,userEvent).Show(true)


    def sendMessageBox(self, jid):
        t = MessageFrame(self, NewId(), jid)
        t.Show(true)
        t.textctrl.SetFocus()

    def setColorForJid(self, jid):
        if(self.list):
            self.list.setColorForJid(jid)

    def removeUser(self, jid):
        if(self.list):
            self.list.removeUser(jid)

    def addUser(self, jid):
        if(self.list):
            self.list.addUser(jid)

    def updateUser(self, jid, nick):
        if(self.list):
            self.list.updateUser(jid, nick)

    def hasUser(self, jid):
        if(self.list):
            return self.list.hasUser(jid)
        return 0

    def IsEmpty(self):
        if(self.list):
            return self.list.IsEmpty()
        return 1




class GUIFrame(wxFrame):
    def __init__(self, parent, ID, title):
        global imcom, eventsQueueHash
        wxFrame.__init__(self, parent, ID, title,
                         wxDefaultPosition, wxSize(200, 150))

        self.splitter = wxSplitterWindow(self, -1,wxPoint(0,0),
                                         wxSize(200,150))
        self.splitter.SetSplitMode(wxSPLIT_HORIZONTAL)
        self.box = wxBoxSizer(wxVERTICAL)
        self.gui = GUI(self.splitter,"MyRoster")
        self.others = GUI(self.splitter,"Others")
        self.gb = GUIbackend.GUIBackend(self)
        imcom = self.gb.imcom
        self.started = 0
        eventsQueueHash = {} # for incoming events from users
        self.splitter.SetMinimumPaneSize(20)
        self.splitter.SplitHorizontally(self.gui, self.others)
        self.splitter.Unsplit(self.others)
        self.others.Show(false)
        #self.box.Add(self.gui, 1, wxEXPAND)
        #self.box.Add(self.others, 0, wxEXPAND|wxTOP, 4)
        self.box.Add(self.splitter, 1, wxEXPAND)
        self.box.Add(wxButton(self, wxID_OK, " IMCom "),
                0, wxALL | wxEXPAND, 1)
        self.SetAutoLayout(true)
        self.SetSizer(self.box)
        self.box.Fit(self)
        self.SetSize(wxSize(64,192))
        self.Show(true)
        EVT_BUTTON(self, wxID_OK, self.OnClick)

        self.mainmenu = wxMenu("IMCom Menu",0)
        self.mainmenu.Append(0, "Online")
        self.mainmenu.Append(1, "Chat")
        self.mainmenu.Append(2, "Away")
        self.mainmenu.Append(3, "XA")
        self.mainmenu.Append(5, "Add User")
        self.mainmenu.Append(4, "Quit")
        EVT_MENU(self, 0, self.SetOnline)
        EVT_MENU(self, 1, self.SetChat)
        EVT_MENU(self, 2, self.SetAway)
        EVT_MENU(self, 3, self.SetXA)
        EVT_MENU(self, 5, self.AddUser)
        EVT_MENU(self, 4, self.Quit)
        EVT_CLOSE(self, self.OnClose)

    def refit(self):
        self.box.Layout()

    def login(self):
        self.gb.sendLogin("floobin.cx", 5222, "test2",
                      "2gwen2", "testing gui")
        imcom.setDebug(1)

    def IMComEventHandler(self, ie):
        print "Received an ", ie.type, " containing data: ", ie.data
        if(ie.type == GUIbackend.IM_GrabRoster):
            # create the tree.
            self.gui.createList()
            self.started = 1
            return
        if(ie.type == GUIbackend.IM_PresenceUpdate):
            data = ie.data
            self.gui.setColorForJid(data['who'])
            # they're not in our list. This is a problem.
            return
        if(ie.type == GUIbackend.IM_VCard):
            # create a Info view
            data = ie.data
            self.ShowUserInfo(data)
            return
        if(ie.type == GUIbackend.IM_MessageReceive):
            data = ie.data

            # append to the EventQueueHash
            ddate = None
            ttime = None
            if(data['delay'] != None):
                delay = data['delay']
                mytup = (int(delay[:4]),int(delay[4:6]),int(delay[6:8]),
                         int(delay[9:11]),int(delay[12:14]),
                         int(delay[15:17]),0,0,0)
                thet = time.localtime( time.mktime(mytup) - time.timezone)

                ttime = time.strftime('%H:%M:%S',thet)
                ddate = time.strftime('%m/%d/%Y',thet)
            else:
                ttime = getTime()
                ddate = getDate()
            if(eventsQueueHash.has_key(data['ffrom'])):
                #they already have events pending, append this one
                eventsQueueHash[data['ffrom']].append( (ie.type,
                                                    data['body'],
                                                    data['thread'],
                                                    ddate,
                                                    ttime))
            else:
                # no events pending from this person, start it.
                eventsQueueHash[data['ffrom']] = [(ie.type,
                                                   data['body'],
                                                   data['thread'],
                                                   ddate,
                                                   ttime)]
            print eventsQueueHash
            if(imcom.jidhash.has_key(data['ffrom'])):
                # Change the color of the User so we realize something changed.
                self.gui.setColorForJid(data['ffrom'])
            else:
                print "Self.started is: ", self.started
                if(0 == self.others.hasUser(data['ffrom']) and
                   self.started == 1):
                    self.others.addUser(data['ffrom'])
                if(not self.splitter.IsSplit() and self.started == 1):
                    self.others.Show(true)
                    self.splitter.SplitHorizontally(self.gui, self.others)
                    #self.refit()
                self.others.setColorForJid(data['ffrom'])
            return
        if(ie.type == GUIbackend.IM_Unsubscribed):
            jid = ie.data['ffrom']
            self.gui.removeUser(jid)
            return
        if(ie.type == GUIbackend.IM_Subscribed):
            jid = ie.data['ffrom']
            if(self.others.hasUser(jid)):
                self.others.removeUser(jid)
                if(self.others.IsEmpty()):
                    self.splitter.Unsplit(self.others)
            if(not self.gui.hasUser(jid)):
                self.gui.addUser(jid)
            return
        if(ie.type == GUIbackend.IM_RosterUpdate):
            if(ie.data['success']):
                self.gui.updateUser(ie.data['jid'], ie.data['name'])
            return
        if(ie.type == GUIbackend.IM_RosterUpdateCheck):
            #if(ie.data['successful']):
            #    self.gui.updateUser(ie.data['jid'], imcom.jidhash[jid])
            return

    def ShowUserInfo(self, data):
        jid = data['ffrom']
        nick = imcom.jidhash[jid]
        status = imcom.preshash[jid][1]
        reason = imcom.preshash[jid][2]
        win = wxFrame(self, -1, "Information on " + nick,
                      wxDefaultPosition, wxSize(400,300))
        #box = wxBoxSizer(wxVERTICAL)
        #hbox = wxBoxSizer(wxHORIZONTAL)
        grid = wxFlexGridSizer(7,2,4,4)

        txt1 = wxStaticText(win, -1, "JID: ")
        txt2 = wxStaticText(win, -1, "First Name: ")
        txt3 = wxStaticText(win, -1, "Last Name: ")
        txt4 = wxStaticText(win, -1, "Nick Name: ")
        txt5 = wxStaticText(win, -1, "Status: ")
        txt6 = wxStaticText(win, -1, "Reason: ")
        edt1 = wxTextCtrl(win, -1, jid, style=wxTE_READONLY)
        edt2 = wxTextCtrl(win, -1, data['given'], style=wxTE_READONLY)
        edt3 = wxTextCtrl(win, -1, data['family'], style=wxTE_READONLY)
        edt4 = wxTextCtrl(win, -1, data['nickname'], style=wxTE_READONLY)
        edt5 = wxTextCtrl(win, -1, status, style=wxTE_READONLY)
        edt6 = wxTextCtrl(win, -1, reason, style=wxTE_READONLY)
        grid.Add(txt1,0,wxALL, 10)
        grid.Add(edt1,1,wxALL|wxEXPAND, 10)
        grid.Add(txt2,0,wxALL, 10)
        grid.Add(edt2,1,wxALL|wxEXPAND, 10)
        grid.Add(txt3,0,wxALL, 10)
        grid.Add(edt3,1,wxALL|wxEXPAND, 10)
        grid.Add(txt4,0,wxALL, 10)
        grid.Add(edt4,1,wxALL|wxEXPAND, 10)
        grid.Add(txt5,0,wxALL, 10)
        grid.Add(edt5,1,wxALL|wxEXPAND, 10)
        grid.Add(txt6,0,wxALL, 10)
        grid.Add(edt6,1,wxALL|wxEXPAND, 10)
        grid.Add(wxButton(win, -1, "Hello"),0,wxALL,10)
        grid.Add(wxButton(win, -1, "Bob"),1,wxALL|wxEXPAND,10)
        grid.AddGrowableCol(1)

        win.SetAutoLayout(true)
        win.SetSizer(grid)
        grid.Fit(win)
        win.Show(true)


    def OnClick(self, event):
        # popup a menu and do stuff..
        self.PopupMenu(self.mainmenu,
                       event.GetEventObject().GetPosition())

    def GetReason(self):
        dlg = wxTextEntryDialog(self,
                                'Enter a reason:',
                                'Reason', "I'm too lazy..")
        if dlg.ShowModal() == wxID_OK:
            return dlg.GetValue()
        return None

    def SetOnline(self, event):
        imcom.sendOnline()

    def SetChat(self, event):
        imcom.sendChat()

    def SetAway(self, event):
        reason = self.GetReason()
        if reason:
            imcom.sendAway(reason)

    def SetXA(self, event):
        reason = self.GetReason()
        if reason:
            imcom.sendXA(reason)

    def AddUser(self, event):
        a = AddUserFrame(self, NewId(), "", "")
        a.Show(true)

    def OnClose(self, event):
        self.mainmenu.Destroy()
        self.gui.finalize()
        self.others.finalize()
        event.Skip()

    def Quit(self, event):
        imcom.disconnect()
        self.Close(true)


class GUIApp(wxApp):
    def OnInit(self):
        self.frame = GUIFrame(None, -1, "Dummy")
        self.SetTopWindow(self.frame)
        self.frame.login()
        #a = wxToolTip("Dummy")
        #a.Enable(true)
        #a.SetDelay(200)
        #a.Destroy()
        #a = None
        return true


if __name__ == "__main__":
    app = GUIApp(0)
    app.MainLoop()
